/*
Created on August 8, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This is the main file for the CalcChu project.  The code for the project is
organized into various smaller components in individual files.  There are
implicit abstraction barriers that are voluntarily observed.  Object-oriented
influences ought to be apparent.

This file contains the starting menu and the code that displays the
instructions.

Profiling suggests that in bad cases (i.e. lots of critters), in updateLinkGame,
a function that is called repeatedly in order to run a game, about two-thirds of
the time is spent in drawLinkGame, and about half of this time is spent in
drawCritters.

The program was compiled using "tigcc -O2 -pack zcalcchu -Wall
-Wno-char-subscripts -Wno-comment calcchu.c extgraph.a", on TIGCC with version
string "tigcc version 1.2.9 built for TIGCC/*nix version 0.95 r3".  It uses the
ExtGraph library (http://tict.ticalc.org), version 2 beta 3.

Version history:

v1.00: Initial release

Program blurb:

This is CalcChu v1.00, a two-player linked game based on "ChuChu Rocket!"  It is
a simple and addictive game.  Cats and mice roam the game board, following
predictable paths.  The goal is to place arrows in order to direct mice to your
goal and cats to your opponent's goal.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#define USE_TI89
#define USE_TI92PLUS
#define USE_V200
#define OPTIMIZE_ROM_CALLS
#define SAVE_SCREEN

#define COMMENT_PROGRAM_NAME "CalcChu"
#define COMMENT_VERSION_STRING "1.00"
#define COMMENT_VERSION_NUMBER 1,0,0,0
#define COMMENT_AUTHORS "William Jacobs"

//Looks like a cat that is chasing a mouse
#define COMMENT_BW_ICON \
	{0x4000, 0x2108, 0x939C, 0x46F6, 0x276E, 0x05FA, 0x07FE, 0x06F6, \
	 0x030C, 0x01F8, 0x0000, 0x00F8, 0xE1FE, 0x07FF, 0xE9FE, 0x00F8}
#define COMMENT_GRAY_ICON \
	{0x0000, 0x0108, 0x039C, 0x06F6, 0x076E, 0x05FA, 0x07FE, 0x06F6, \
	 0x030C, 0x01F8, 0x0000, 0x00F8, 0x01FE, 0x07FF, 0x09FE, 0x00F8}, \
	{0x4000, 0x2108, 0x939C, 0x46F6, 0x276E, 0x05FA, 0x07FE, 0x06F6, \
	 0x030C, 0x01F8, 0x0000, 0x00F8, 0xE1FE, 0x07FF, 0xE9FE, 0x00F8}

#include <tigcclib.h>

/*
typedef struct {
	long t;
	short n;
} Timer;

#define START(x) (x).t -= time()
#define STOP(x) (x).t += time(); ((x).n)++
#define AVG(x) (((float)((x).t)) / ((float)((x).n)))

Timer T1;
Timer T2;
Timer T3;
Timer T4;
Timer T5;
*/

#include "draw.h"
#include "key.h"
#include "linkmatch.h"
#include "menu.h"
#include "time.h"

//Draws the first screen of the instructions
void drawInstructions1() {
	drawMessage2("CalcChu is a two-player game that is easy to learn and play. "
				 "The object is to direct mice to your goal and direct cats to "
				 "your opponent's goal.");
}

//Draws the second screen of the instructions
void drawInstructions2() {
	//By fortunate happenstance, offsetting the board by a 16 pixels
	//horizontally (a multiple of 8) places it in the center of the screen
	//(although it is has one more pixel to the right of it than to the left,
	//breaking my convention of having one more pixel to the left)
	
	char buffer[LCD_SIZE];
	//30 = 240 / 8
	short offset = 13 * 30 + 2;
	Board *board = loadBuiltInBoard(17);
	
	clearDrawBuffer(buffer);
	drawBoard(buffer + offset, board);
	drawRocket(buffer + offset, board->rocketAPos, 1);
	drawRocket(buffer + offset, board->rocketBPos, 0);
	freeBoard(board);
	free(board);
	
	setDrawBuffer(buffer);
	setFont6x8();
	drawCenter(2, "A CalcChu game board:");
	PortRestore();
	draw89Size(buffer);
}

//Draws the third screen of the instructions
void drawInstructions3() {
	char buffer[LCD_SIZE];
	
	clearAndSetDrawBuffer(buffer);
	setFont6x8();
	drawCenter(2, "These are the different");
	drawCenter(11, "game objects:");
	
	setFont4x6();
	Sprite8_OR(2, 23, 7, P1_ROCKET, buffer);
	Sprite8_OR(11, 23, 7, P2_ROCKET, buffer);
	drawStrNormal(21, 24, "Goals for the white and black players");
	
	Sprite16_OR(2, 37, 5, MOUSE_EAST, buffer);
	drawStrNormal(16, 34, "A mouse, worth one point when it enters");
	drawStrNormal(16, 41, "your goal");
	
	Sprite16_OR(2, 52, 13, CAT_SOUTH, buffer);
	drawStrNormal(19, 53, "A cat, which subtracts one third of your");
	drawStrNormal(19, 60, "score when it enters your goal");
	
	Sprite16_OR(2, 72, 9, SUPER_MOUSE_PIC, buffer);
	drawStrNormal(17, 71, "A super mouse, worth fifty points when it");
	drawStrNormal(17, 78, "enters your goal");
	
	setFont6x8();
	drawCenter(89, "(more on next screen)");
	
	PortRestore();
	draw89Size(buffer);
}

//Draws the fourth screen of the instructions
void drawInstructions4() {
	char buffer[LCD_SIZE];
	
	clearAndSetDrawBuffer(buffer);
	setFont6x8();
	drawCenter(2, "These are the different");
	drawCenter(11, "game objects:");
	
	setFont4x6();
	Sprite8_OR(2, 27, 11, SPECIAL_MOUSE_PIC, buffer);
	drawStrNormal(12, 23, "A special mouse. Starts a special event");
	drawStrNormal(12, 30, "when it enters a goal. There are 10 special");
	drawStrNormal(12, 37, "events; learn them as you go.");
	
	Sprite8_OR(2, 47, 5, HATCH_EAST, buffer);
	drawStrNormal(10, 47, "A hatch. Mice and cats appear out of here.");
	
	Sprite16_OR(2, 57, 13, PIT, buffer);
	drawStrNormal(18, 58, "A pit. Mice and cats disappear when they");
	drawStrNormal(18, 65, "hit one.");
	
	Sprite16_OR(2, 75, 11, P1_ARROW_EAST, buffer);
	drawStrNormal(16, 75, "An arrow. Mice and cats go the direction");
	drawStrNormal(16, 82, "it is pointing when they hit one.");
	
	PortRestore();
	draw89Size(buffer);
}

//Draws the fifth screen of the instructions
void drawInstructions5() {
	drawMessage2("The mice and cats follow predictable paths. They go straight "
				 "unless they hit an arrow or wall.");
}

//Draws the sixth screen of the instructions
void drawInstructions6() {
	drawMessage2("When they hit a wall head-on, they turn right. When they hit "
				 "a corner, they follow it. When they hit a dead end, they "
				 "turn around.");
}

//Draws the seventh screen of the instructions
void drawInstructions7() {
	drawMessage2("Each player has three arrows to place on the board to direct "
				 "the mice and cats. You can place an arrow by selecting a "
				 "space using the arrow keys and by pressing 8, 6, 2, or 4.");
}

//Draws the eighth screen of the instructions
void drawInstructions8() {
	drawMessage2("On the TI-89 and TI-89 Titanium, you can also place an arrow "
				 "by pressing MODE, Z, (, or X. Whoever scores the most points "
				 "when time runs out wins the game!");
}

#define NUM_INSTRUCTIONS_SCREENS 8

//Shows the game instructions, by calling the drawInstructionsX methods and
//responding to key presses
void showInstructions() {
	short i;
	unsigned short key;
	static void (*instructions[NUM_INSTRUCTIONS_SCREENS])(void) =
		{drawInstructions1,
		 drawInstructions2,
		 drawInstructions3,
		 drawInstructions4,
		 drawInstructions5,
		 drawInstructions6,
		 drawInstructions7,
		 drawInstructions8};
	
	i = 0;
	do {
		instructions[i]();
		idle();
		while(1) {
			key = nextKey();
			if (key == KEY_ESC) {
				i = -1;
				break;
			}
			else if (key == KEY_ENTER || key == KEY_RIGHT || key == KEY_DOWN) {
				i++;
				break;
			}
			else if (key == KEY_LEFT || key == KEY_UP) {
				i--;
				break;
			}
		}
	}
	while (i >= 0 && i < NUM_INSTRUCTIONS_SCREENS);
}

/*
//Displays the build-in boards in order.  Not intended for use in the final
//program.
void showBoards() {
	short i;
	char buffer[LCD_SIZE];
	short offset = 13 * 30 + 2;
	char s[10];
	
	for(i = 0; i < NUM_BUILT_IN_BOARDS; i++) {
		Board *board = loadBuiltInBoard(i);
		
		clearDrawBuffer(buffer);
		drawBoard(buffer + offset, board);
		drawRocket(buffer + offset, board->rocketAPos, 1);
		drawRocket(buffer + offset, board->rocketBPos, 0);
		freeBoard(board);
		free(board);
		
		setDrawBuffer(buffer);
		setFont6x8();
		sprintf(s, "%d", i);
		drawCenter(2, s);
		PortRestore();
		draw89Size(buffer);
		while(!nextKey());
	}
}
*/

void _main(void) {
	short *results;
	short s;
	Menu *menu;
	
	#if 0
	T1.t = 0; T1.n = 0;
	T2.t = 0; T2.n = 0;
	T3.t = 0; T3.n = 0;
	T4.t = 0; T4.n = 0;
	T5.t = 0; T5.n = 0;
	#endif
	
	//The amount of heap space available is a heuristic for the ability to
	//allocate the necessary chunks of memory from the heap.  Approximately 5000
	//bytes should be used by the game at peak heap usage (~2500 for caching the
	//drawn board, ~1500 for other game data, and I'm guessing < ~1000 for
	//static variables (those are in the heap, right?) and temporary usage).
	if (HeapAvail() >= 8000) {
		//Must be called before initKeysAndTimer messes with the timer;
		//otherwise, the RNG is always seeded with the same value
		randomize();
		
		initKeysAndTimer();
		
		TRY
		
		SetCurClip(&(SCR_RECT){{0, 0, 239, 127}});
		
		menu = newMenu(1);
		addMenuItem(menu, newMenuItem("CalcChu, v1.00", 0, F_8x10), 8);
		addMenuItem(menu, newMenuItem("Create game", 1, F_6x8), 1);
		addMenuItem(menu, newMenuItem("Join game", 1, F_6x8), 1);
		addMenuItem(menu, newMenuItem("Instructions / Controls", 1, F_6x8), 1);
		addMenuItem(menu, newMenuItem("Exit", 1, F_6x8), 0);
		
		while(1) {
			results = presentMenu(menu, 0);
			if (results != NULL) {
				s = results[0];
				free(results);
				
				if (s == 3)
					break;
				
				switch (s) {
					case 0:
						hostLinkMatch();
						break;
					case 1:
						joinLinkMatch();
						break;
					case 2:
						showInstructions();
				}
			}
			else
				break;
		}
		
		freeMenu(menu);
		free(menu);
		
		FINALLY
		
		cleanupKeysAndTimer();
		
		ENDFINAL
	}
	else {
		drawMessage2("Sorry, you don't have enough memory to play "
					 "CalcChu v1.00");
		idle();
		flushKeyQueue();
		while (!nextKey());
	}
}














