/*
Created on August 9, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file contains the data model for the game and the board, as well as
functions to advance the state of the game board and perform other
manipulations.  The game advances in discrete, atomic intervals refered to as
"ticks".

There is a good amount of data duplication in the representation of a game.  The
program relies on careful maintenance of certain invariants in the
representation.  In particular, Game.boardData must be kept consistent with the
other game data.

Each built-in board is represented compactly as a sequence of BOARD_SIZE bytes,
allowing for easy loading and transferring of boards.

To keep the games displayed on linked calculators identical, a random number
generator seeded with the same value is used on each calculator, the usage of
which is carefully controlled.  Thus, seemingly innocent changes to some
algorithms may result in incompatibility with alternative versions of the game.
The rules for executing the game are as follows, with uses of the random number
generator indicated:

- At the beginning of the game, select a random number using rngRandom(rng, 2).
If this is 0, set player 1's rocket to be the rocketA field in Board and player
2's rocket to be the other one; if the result is 1, set the players' rockets to
be the reverse of this.

- In the body of hatchTick, hatch relevant critters if the number of critters is
not already equal to the maximum as follows: compute the relevant type to hatch
for each of the hatches, in the order they appear in the Board.  Until the
number of critters is hatched is less than the maximum number of critters minus
the number of critters in the game, do the following: select a random element
among the types (excluding those that are NO_CRITTER) by calling rngRandom(rng,
number_of_types).  Remove this type, and move all subsequent types so that they
are one element earlier.

When a critter type to be hatched is randomly selected by using probabilities
for whether a critter will be generated and that each critter type will be
selected, when a critter is generated, it is done as follows: call
rngRandom(rng, 10000), and hatch something if the result is less than or equal
to 10000 * rate_probability.  Call rngRandom again, and store the result.  If it
is less than or equal to 10000 * cat_probability, hatch a cat.  Otherwise, if it
is less than or equal to 10000 * (cat_probability + super_mouse_probability),
hatch a super mouse.  Otherwise, if it is less than or equal to 10000 *
(cat_probability + super_mouse_probability + special_mouse_probability), hatch a
special mouse.  Otherwise, hatch a normal mouse.  The exception to this is when
one or more of the probabilities is 0, in which case the appropriate branches
are excluded.  Note that obviously, cat_probability + super_mouse_probability +
special_mouse_probability should be at most 1, and rate_probability should be at
most 1.

- After hatching the relevant critter(s), destroy each mouse for which the space
to which it is closest (or one of the two spaces to which it is closest, if it
is equidistant from two spaces) is the same as the space to which a cat is
closest.  Distance from a space takes only the critter's portion field into
account.

- In the body of advanceMiceTick, move mice as appropriate, destroying them if
they are "near" a cat, as described above.  If a special mouse entered one of
the two rockets, indicate a special event for the player whose hatch it entered.
If a special mouse entered each of the two rockets, call rngRandom(rng, 2), and
indicate a special event for player 1 if the result is 0 and for player 2 if the
result is 1.

- The body of advanceTick hatches critters, then moves cats, then moves mice.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef GAME_H_INCLUDED
#define GAME_H_INCLUDED

#include "bitstream.h"
#include "rng.h"

//The dimensions of the board.  Note that these values may be taken to be fixed;
//they are provided for more readable code.
#define WIDTH 9
#define HEIGHT 6
#define WIDTH_TIMES_HEIGHT 54

//The super mouse is the one worth SUPER_SCORE points, and the special mouse is
//the one that generates a special event.  Note that MOUSE, SUPER_MOUSE, and
//SPECIAL_MOUSE must be 0, 1, and 2 (though not necessarily in that order) in
//order for critter "stacking" to work in advanceMiceTick.  The set of mouse
//types may be taken to be fixed.
enum CritterType
	{MOUSE = 0, SUPER_MOUSE = 1, SPECIAL_MOUSE = 2, NO_CRITTER, CAT};
enum Directions {NORTH = 0, EAST = 1, SOUTH = 2, WEST = 3, NO_DIRECTION};
//Used in SpaceData to indicate what is occupying a particular space
enum SpaceType
	{SPACE_NOTHING, SPACE_PIT, SPACE_ROCKET1, SPACE_ROCKET2, SPACE_HATCH};

//The maximum number of arrows that can be on the board for any one player
#define MAX_ARROWS 3
//The amount by which a cat advances in a single game tick, as a portion of
//MAX_PORTION.  It must be 2; it is provided for readability.
#define CAT_ADVANCE 2
//The amount by which a mouse advances in a single game tick, as a portion of
//MAX_PORTION.  It must be 3; it is provided for readability.
#define MOUSE_ADVANCE 3
//Must be 6; it is provided for readability
#define MAX_PORTION 6
//The score for capturing a super mouse
#define SUPER_SCORE 50
//The maximum score a player can have at any time.  May be at most 999, unless
//the score drawing function is changed, and ought to be 999.
#define MAX_SCORE 999
//The number of bytes used to store a board.  The amount is the same for every
//board.
#define BOARD_SIZE 27
//The maximum number of critters that may be in play at once.  Right now, this
//cannot be more than 127.  Only the code in advanceCatsTick and advanceMiceTick
//that tries to stack multiple critters into one Critter object depends on this
//being less than 128, so to exceed this limit, one need only alter this code to
//accept larger values.  Further, the fact that the numCritters field of the
//Critter object is a unsigned char limits this to 255, but changing this to a
//short would be trivial.
#define MAX_CRITTERS 100
//The maximum number of hatches a level may contain.  All built-in boards
//contain 2 hatches, but the program is capable of loading boards with 0 to 5
//hatches.  This allows for future expansion, e.g. a board editor.
#define MAX_HATCHES 5

//The probabilities used for hatching under normal circumstances.  Some
//functions in timedgame.h rely on these values' not being excessively high, as
//they are modified using multiplication, and no pertinent probability should
//exceed 1.

//10000 times the probability that a hatch will generate a cat under normal
//circumstances
#define NORMAL_CAT_PROBABILITY_10000 300
//10000 times the probability that a hatch will generate a super mouse under
//normal circumstances
#define NORMAL_SUPER_PROBABILITY_10000 200
//10000 times the probability that a hatch will generate a special mouse under
//normal circumstances
#define NORMAL_SPECIAL_PROBABILITY_10000 400
//10000 times the probability that a hatch will hatch something at each tick
//under normal circumstances
#define NORMAL_HATCH_RATE_10000 800

//The value in Board.walls that corresponds to walls on all four sides of a
//space
#define WALLS_ALL 15

//Represents a cat or mouse, or multiple cats or mice with the same properties.
//This sort of stacking makes for faster drawing in some of the more common bad
//cases, and possibly for faster updating.  Its internal representation is
//transparent and may be accessed outside of this file.
//5 bytes
typedef struct {
	//The number of critters with the same set of properties that are
	//represented by this Critter object (at least 1).  It is not required that
	//there exist no other Critter object representing critter(s) with the same
	//properties in the game.  If 0, then this object does not represent a
	//critter that is actually in play, but rather one that is merely occupying
	//space.  In that case, no guarantees are made regarding the other fields.
	//Normally, this field would be a short, as per the convention of not
	//sweating that extra byte.  It is an unsigned char instead based on the
	//(possibly false) guess that comparing it with 0 is faster than comparing a
	//short with 0, and since this comparison seems particularly important for
	//speed efficiency.
	unsigned char numCritters;
	//The current position of this, represented as y * WIDTH + x
	char pos;
	//The extent to which this has advanced to its next position, as a fraction
	//of MAX_PORTION.  When 0, it indicates that the critter has just arrived at
	//its current position.  It must be 0 or 3 for mice and 0, 2, or 4 for cats.
	char portion;
	//The type of this: CAT, MOUSE, SUPER_MOUSE, or SPECIAL_MOUSE.  It may not
	//be NO_CRITTER.
	char type;
	//The direction in which this is currently traveling
	char dir;
} Critter;

//Represents a hatch from which cats and mice appear.  Its internal
//representation is transparent and may be accessed outside of this file.
//3 bytes
typedef struct {
	//The position of this, represented as y * WIDTH + x
	char pos;
	//The direction which creatures generated from this appear facing.  May be
	//facing the wall.
	char displayDir;
	//The direction which creatures generated from this appear facing.  It is
	//the same as displayDir, except that it is adjusted so as not to face the
	//wall.
	char dir;
} Hatch;

//Represents the walls, pits, hatches, and rockets that comprise the board.
//Should not be altered after it is loaded.  Its internal representation is
//transparent and may be accessed outside of this file.
//140 bytes
typedef struct {
	//The walls in this.  Each element stores the adjacent walls for one space.
	//The last four bits represent whether there is a wall in each of the
	//directions, so that (wall >> dir) & 1 == 1 iff there is a wall in
	//direction dir.  Of course, it is in row-major order.
	char walls[WIDTH_TIMES_HEIGHT];
	//The pits in this.  Each element stores whether a space contains a pit.  Of
	//course, it is in row-major order.
	char pits[WIDTH_TIMES_HEIGHT];
	//The hatches in this.  Note that only the first numHatches elements have
	//meaningful data.
	Hatch hatches[MAX_HATCHES];
	//The number of hatches in this
	short numHatches;
	//The position of the first rocket, which receives mice and cats,
	//represented as y * WIDTH + x.  Note that it may be either player 1's
	//rocket or player 2's rocket.
	char rocketAPos;
	//The position of the second rocket, which receives mice and cats,
	//represented as y * WIDTH + x.  Note that it may be either player 1's
	//rocket or player 2's rocket.
	char rocketBPos;
} Board;

//Represents an arrow positioned by a player.  Its internal representation is
//transparent and may be accessed outside of this file.
//4 bytes
typedef struct {
	//Whether this Arrow object represents a arrow that is actually in play, as
	//opposed to merely occupying space.  Note that no guarantees are made
	//regarding the other fields if this is false.
	char exists;
	//The position of this, represented as y * WIDTH + x
	char pos;
	//The direction which this is pointing
	char dir;
	//Whether the arrow is small because it has been partially eaten by a cat
	char isSmall;
} Arrow;

//Data for a single space in the board.  Its internal representation is
//transparent and may be accessed outside of this file.
//4 bytes
typedef struct {
	//The number of stacks of cats (that is, the number of Critter objects) that
	//are closer to the space than any other
	short numCats;
	//The type of the space, e.g. SPACE_NONE, SPACE_PIT
	char type;
	//The direction of the arrow on this, or NO_DIRECTION
	char arrowDir;
} SpaceData;

//Represents information regarding a game in progress.  Its internal
//representation is transparent and may be accessed outside of this file.
//798 bytes
typedef struct {
	//The critters in play (excluding for those for which critter.numCritters ==
	//0)
	Critter critters[MAX_CRITTERS];
	//The total number of critters in play
	short numCritters;
	//The game board
	Board *board;
	//The arrows in play for player 1 (excluding for those for which
	//!arrow.exists)
	Arrow p1Arrows[MAX_ARROWS];
	//The arrows in play for player 1 (excluding for those for which
	//!arrow.exists)
	Arrow p2Arrows[MAX_ARROWS];
	//Player 1's score
	short score1;
	//Player 2's score
	short score2;
	//Data regarding each space on the board.  Anything that modifies this game
	//must ensure that boardData remains consistent with the contents of this
	//game.
	SpaceData boardData[WIDTH_TIMES_HEIGHT];
	//The random number generator used by this
	RNG *rng;
} Game;



//The generator function used for hatches under normal circumstances
char normalGenerator(RNG *rng) {
	short r;
	
	if (rngRandom(rng, 10000) <= NORMAL_HATCH_RATE_10000) {
		r = rngRandom(rng, 10000);
		if (r <= NORMAL_CAT_PROBABILITY_10000)
			return CAT;
		else if (r <= NORMAL_CAT_PROBABILITY_10000 +
					  NORMAL_SUPER_PROBABILITY_10000)
			return SUPER_MOUSE;
		else if (r <= NORMAL_CAT_PROBABILITY_10000 +
					  NORMAL_SUPER_PROBABILITY_10000 +
					  NORMAL_SPECIAL_PROBABILITY_10000)
			return SPECIAL_MOUSE;
		else
			return MOUSE;
	}
	else
		return NO_CRITTER;
}

//Creates a critter of the specified type from the specified hatch.  Assumes
//that type is not NO_CRITTER.
Critter *newCritter(const Hatch *hatch, char type) {
	Critter *critter = (Critter *)malloc_throw(sizeof(Critter));
	critter->numCritters = 1;
	critter->pos = hatch->pos;
	critter->portion = 0;
	critter->type = type;
	critter->dir = hatch->dir;
	return critter;
}

//Adds the specified critter to the game, if possible.  Assumes that
//critter->numCritters == 1 and critter->portion == 0.  It does not take charge
//of freeing the critter, but rather copies its contents to some element of
//game->critters.
void addCritter(Game *game, Critter *critter) {
	short i;
	Critter *critter2;
	
	for(i = 0, critter2 = game->critters; i < MAX_CRITTERS; i++, critter2++) {
		if (!(critter2->numCritters)) {
			memcpy(critter2, critter, sizeof(Critter));
			if (critter->type == CAT)
				game->boardData[critter->pos].numCats++;
			(game->numCritters)++;
			return;
		}
	}
}

//Returns the new direction that a critter should move, given its current
//direction and the walls surrounding its current position.  Assumes that walls
//is not WALLS_ALL.
static char newDir(char dir, char walls) {
	return ((const char[64])
	{NORTH, EAST, NORTH, WEST, NORTH, EAST, NORTH, WEST,
			NORTH, EAST, NORTH, SOUTH, NORTH, EAST, NORTH, NO_DIRECTION,
	 EAST, EAST, SOUTH, SOUTH, EAST, EAST, NORTH, WEST,
			EAST, EAST, SOUTH, SOUTH, EAST, EAST, NORTH, NO_DIRECTION,
	 SOUTH, SOUTH, SOUTH, SOUTH, WEST, WEST, WEST, WEST,
			SOUTH, SOUTH, SOUTH, SOUTH, EAST, EAST, NORTH, NO_DIRECTION,
	 WEST, WEST, WEST, WEST, WEST, WEST, WEST, WEST,
			NORTH, SOUTH, NORTH, SOUTH, NORTH, EAST, NORTH, NO_DIRECTION})
		[(dir << 4) + walls];
}

//Returns the new direction that a critter should move, given its current
//direction, the walls surrounding its current position, and the direction of
//the arrow on its current space (which may be NO_DIRECTION).  Assumes that
//walls is not WALLS_ALL.
static inline char newDirArrow(char dir, char arrow, char walls) {
	return newDir(arrow == NO_DIRECTION ? dir : arrow, walls);
}

//Destroys all mice that are on the same space as a cat
void destroyEatenMice(Game *game) {
	short i;
	Critter *critter;
	
	for(i = 0, critter = game->critters; i < MAX_CRITTERS; i++, critter++) {
		if (critter->numCritters && critter->type != CAT &&
				game->boardData[critter->pos].numCats) {
			game->numCritters -= critter->numCritters;
			critter->numCritters = 0;
		}
	}
}

//Generates all mice and cats appropriate for a single game tick.  Takes a
//generator that returns the type of critter to generate.
void hatchTick(Game *game, char (*generator)(RNG *)) {
	short i;
	short critterNum;
	
	SpaceData *boardData = game->boardData;
	Critter critters[MAX_HATCHES];
	Critter *critter;
	Critter *critter2;
	char type;
	
	if (game->numCritters == MAX_CRITTERS)
		return;
	
	critterNum = 0;
	for(i = 0; i < game->board->numHatches; i++) {
		type = generator(game->rng);
		if (type != NO_CRITTER) {
			critter = newCritter(game->board->hatches + i, type);
			memcpy(critters + critterNum++, critter, sizeof(Critter));
			free(critter);
		}
	}
	
	
	
	//Remove critters from the set of those to be hatched until there is room to
	//ensure there will be no more than MAX_CRITTERS critters
	while (game->numCritters + critterNum > MAX_CRITTERS) {
		i = rngRandom(game->rng, critterNum);
		memmove(critters + i, critters + i + 1,
				(critterNum - i - 1) * sizeof(Critter));
		critterNum--;
	}
	
	critter = critters;
	critter2 = game->critters;
	i = 0;
	while (i < critterNum) {
		if (!critter2->numCritters) {
			memcpy(critter2, critter, sizeof(Critter));
			if (critter->type == CAT)
				boardData[critter->pos].numCats++;
			i++;
			critter++;
		}
		critter2++;
	}
	
	game->numCritters += critterNum;
	
	
	
	//Destroy mice as appropriate
	if (critterNum)
		destroyEatenMice(game);
}

//Returns the space that is one square in the specified direction from the
//specified position
char advancePos(char pos, char dir) {
	switch (dir) {
		case NORTH:
			if (pos < WIDTH)
				return pos + WIDTH_TIMES_HEIGHT - WIDTH;
			else
				return pos - WIDTH;
		case EAST:
			if (++pos % WIDTH)
				return pos;
			else
				return pos - WIDTH;
		case SOUTH:
			if (pos < WIDTH_TIMES_HEIGHT - WIDTH)
				return pos + WIDTH;
			else
				return pos - WIDTH_TIMES_HEIGHT + WIDTH;
		//case WEST:
		default:
			if (pos-- % WIDTH)
				return pos;
			else
				return pos + WIDTH;
	}
}

//Returns the space to which the specified critter is currently moving
char nextPos(Critter *critter) {
	return advancePos(critter->pos, critter->dir);
}

//Returns the resultant score after getting a cat
short scoreCat(short score) {
	return max(score - max(score / 3, 1), 0);
}

//Returns the opposite of the given direction.  (Assumes that dir !=
//NO_DIRECTION.)
char opposite(char dir) {
	switch (dir) {
		case NORTH:
			return SOUTH;
		case EAST:
			return WEST;
		case SOUTH:
			return NORTH;
		//case WEST:
		default:
			return EAST;
	};
}

//Moves the specified cat as appropriate for a single game tick.  Assumes that
//the critter exists and is a cat.
void advanceCatTick(Game *game, Critter *critter) {
	short i;
	
	SpaceData *boardData = game->boardData;
	char pos;
	SpaceData *space;
	SpaceData *nextSpace;
	short *score;
	char arrowDir;
	Arrow *arrow;
	
	pos = critter->pos;
	space = boardData + pos;
	pos = nextPos(critter);
	
	if ((critter->portion += CAT_ADVANCE) >= MAX_PORTION) {
		nextSpace = boardData + pos;
		
		switch (nextSpace->type) {
			case SPACE_ROCKET1:
			case SPACE_ROCKET2:
				score = nextSpace->type == SPACE_ROCKET1 ?
					&(game->score1) : &(game->score2);
				for(i = 0; i < critter->numCritters; i++)
					*score = scoreCat(*score);
				//No break statement
			case SPACE_PIT:
				game->numCritters -= critter->numCritters;
				nextSpace->numCats--;
				critter->numCritters = 0;
				break;
			default:
				critter->portion = 0;
				critter->pos = pos;
				arrowDir = nextSpace->arrowDir;
				//critter->dir = newDirArrow(critter->dir,
				//						   arrowDir,
				//						   game->board->walls[pos]);
				critter->dir =
					newDir(arrowDir == NO_DIRECTION ? critter->dir : arrowDir,
						   game->board->walls[pos]);
				
				if (arrowDir != NO_DIRECTION) {
					//Make the arrow small, or remove it if it is already small
					for(i = 0, arrow = game->p1Arrows;
						i < MAX_ARROWS && (!(arrow->exists) ||
										   arrow->pos != pos);
						i++, arrow++);
					if (i >= MAX_ARROWS)
						//We shouldn't exceed the limit of game->p2Arrows.  This
						//would indicate an inconsistent Game structure
						for(i = 0, arrow = game->p2Arrows;
							!(arrow->exists) || arrow->pos != pos;
							i++, arrow++);
					
					if (arrow->isSmall) {
						arrow->exists = 0;
						nextSpace->arrowDir = NO_DIRECTION;
					}
					else
						arrow->isSmall = 1;
				}
		}
	}
	else if (critter->portion == 4) {
		//The cat is now closer to the next space than to the current one
		space->numCats--;
		boardData[pos].numCats++;
	}
}

//Moves all cats as appropriate for a single game tick
void advanceCatsTick(Game *game) {
	short i;
	//768 = 3 * 256
	char cats[768];
	char *index;
	Critter *critter;
	SpaceData *boardData = game->boardData;
	Critter *critters = game->critters;
	
	memset(cats, -1, 768);
	
	for(i = 0, critter = critters; i < MAX_CRITTERS; i++, critter++) {
		if (critter->numCritters && critter->type == CAT) {
			//This provides an index into cats that uniquely identifies the
			//cat's portion, position, and direction.  Note that some elements
			//of cats are not used.
			index = cats + (((short)critter->portion) << 7) +
				(((short)critter->pos) << 2) + critter->dir;
			if (*index >= 0) {
				if (critters[*index].numCritters)
					critters[*index].numCritters += critter->numCritters;
				else
					game->numCritters -= critter->numCritters;
				
				boardData[critter->portion < 3 ?
						  critter->pos : nextPos(critter)].numCats--;
				critter->numCritters = 0;
			}
			else {
				*index = i;
				advanceCatTick(game, critter);
			}
		}
	}
}

//Moves the specified mouse as appropriate for a single game tick.  Assumes that
//the critter exists and is a mouse.  Returns whether a special event was
//triggered, and if so, the player number for whom it is triggered (1 or 2).
//Note that it is important to advance the cats before the mice, because when
//the mice are moved, those that are on the same space as a cat are destroyed.
char advanceMouseTick(Game *game, Critter *critter) {
	char pos;
	SpaceData *space;
	short *score;
	char destroy;
	short specialEvent = 0;
	
	pos = critter->pos;
	if ((critter->portion += MOUSE_ADVANCE) >= MAX_PORTION) {
		pos = nextPos(critter);
		space = game->boardData + pos;
		
		switch (space->type) {
			case SPACE_ROCKET1:
			case SPACE_ROCKET2:
				if (critter->type != SPECIAL_MOUSE) {
					score = space->type == SPACE_ROCKET1 ?
						&(game->score1) : &(game->score2);
					*score = min(*score + critter->numCritters *
									(critter->type == MOUSE ? 1 : SUPER_SCORE),
								 MAX_SCORE);
				}
				else
					specialEvent = space->type == SPACE_ROCKET1 ? 1 : 2;
				//No break statement
			case SPACE_PIT:
				destroy = 1;
				break;
			default:
				destroy = space->numCats > 0;
				if (!destroy) {
					critter->portion = 0;
					critter->pos = pos;
					critter->dir = newDir(space->arrowDir == NO_DIRECTION ?
											critter->dir : space->arrowDir,
										  game->board->walls[pos]);
				}
		}
	}
	else
		//The mouse is halfway to the next space
		destroy = game->boardData[pos].numCats ||
			game->boardData[nextPos(critter)].numCats;
	
	if (destroy) {
		game->numCritters -= critter->numCritters;
		critter->numCritters = 0;
	}
	
	return specialEvent;
}

//Moves all mice as appropriate for a single game tick.  Returns whether a
//special event was triggered, and if so, the player number for whom it is
//triggered (1 or 2).  Note that it is important to advance the cats before the
//mice, because when the mice are moved, those that are on the same space as a
//cat are destroyed.
char advanceMiceTick(Game *game) {
	short i;
	
	Critter *critter;
	//1536 = 6 * 256
	char mice[1536];
	char *index;
	char specialEventP1 = 0;
	char specialEventP2 = 0;
	Critter *critters = game->critters;
	
	memset(mice, -1, 1536);
	
	for(i = 0, critter = critters; i < MAX_CRITTERS; i++, critter++) {
		if (critter->numCritters && critter->type != CAT) {
			//This provides an index into mice that uniquely identifies the
			//mouse's portion, type, position, and direction.  Note that some
			//elements of mice are not used.
			index = mice + ((((short)(critter->portion + critter->type)) << 8) +
							(((short)critter->pos) << 2) + critter->dir);
			
			if (*index >= 0) {
				if (critters[*index].numCritters)
					critters[*index].numCritters += critter->numCritters;
				else
					game->numCritters -= critter->numCritters;
				critter->numCritters = 0;
			}
			else {
				*index = i;
				switch(advanceMouseTick(game, critter)) {
					case 1:
						specialEventP1 = 1;
						break;
					case 2:
						specialEventP2 = 1;
				}
			}
		}
	}
	
	if (specialEventP1) {
		if (specialEventP2)
			return rngRandom(game->rng, 2) + 1;
		else
			return 1;
	}
	else if (specialEventP2)
		return 2;
	else
		return 0;
}

//Advances the position by one game tick.  Returns whether a special event was
//triggered, and if so, the player number for whom it is triggered (1 or 2).
//Note that it is important to advance the cats before the mice, because when
//the mice are moved, those that are on the same space as a cat are destroyed.
char advanceTick(Game *game) {
	hatchTick(game, normalGenerator);
	advanceCatsTick(game);
	return advanceMiceTick(game);
}

//Destroys all critters in the game
void destroyCritters(Game *game) {
	short i;
	
	//Critter *critter;
	//for(i = 0, critter = game->critters; i < MAX_CRITTERS; i++, critter++)
	//	critter->numCritters = 0;
	
	memset(game->critters, 0, MAX_CRITTERS * sizeof(Critter));
	
	for(i = 0; i < WIDTH_TIMES_HEIGHT; i++)
		game->boardData[i].numCats = 0;
	
	game->numCritters = 0;
}

//Returns the number of cats currently in play
short numCats(Game *game) {
	short i;
	short num = 0;
	Critter *critter;
	
	for(i = 0, critter = game->critters; i < MAX_CRITTERS; i++, critter++) {
		if (critter->type == CAT)
			num += critter->numCritters;
	}
	return num;
}

//The number of built-in boards.  Must be at least 2.
#define NUM_BUILT_IN_BOARDS 25

//Returns the raw data for the built-in board with the specified index.  Assumes
//that 0 <= index < BUILT_IN_BOARDS.  The data can be converted to a Board
//object by calling loadBoard.  These boards were constructed using a Java
//program written for the purpose.  The boards are designed so that none of the
//following conditions can occur by the placement of one player's arrows,
//regardless of the placement of the other player's arrows: all of the critters
//from one hatch go to a particular rocket, or none of the critters reach one
//rocket. Hopefully, I did not accidentally design a board that does not meet
//this intention.
unsigned char *builtInBoard(short index) {
	static unsigned char levels[NUM_BUILT_IN_BOARDS][BOARD_SIZE] =
	{{170, 170, 128, 0, 4, 2, 33, 36, 0, 16, 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 15,
	  160, 162, 187, 243, 243, 240},
	 {235, 170, 144, 0, 212, 0, 33, 0, 0, 116, 0, 24, 1, 0, 0, 0, 0, 0, 0, 0,
	  19, 138, 136, 211, 243, 243, 240},
	 {234, 170, 128, 0, 4, 0, 1, 213, 16, 117, 68, 24, 34, 0, 0, 0, 0, 0, 0, 0,
	  9, 138, 97, 243, 243, 243, 240},
	 {74, 34, 144, 0, 4, 0, 1, 0, 0, 64, 0, 16, 0, 0, 8, 4, 2, 1, 0, 128, 70,
	  196, 146, 83, 243, 243, 240},
	 {234, 184, 146, 72, 4, 64, 5, 9, 8, 64, 0, 17, 2, 0, 0, 0, 0, 0, 0, 0, 16,
	  173, 197, 39, 243, 243, 240},
	 {203, 44, 128, 0, 4, 16, 67, 44, 178, 0, 0, 16, 65, 0, 0, 0, 0, 0, 0, 0,
	  10, 148, 91, 35, 243, 243, 240},
	 {170, 170, 128, 0, 0, 136, 128, 136, 136, 8, 136, 0, 0, 0, 0, 0, 0, 0, 0,
	  0, 14, 165, 181, 39, 243, 243, 240},
	 {3, 164, 0, 0, 12, 0, 9, 0, 0, 128, 0, 128, 65, 0, 0, 0, 0, 0, 0, 0, 46,
	  208, 193, 143, 243, 243, 240},
	 {226, 162, 144, 0, 4, 221, 209, 34, 32, 119, 119, 88, 136, 128, 0, 0, 0, 0,
	  0, 0, 2, 26, 66, 195, 243, 243, 240},
	 {234, 186, 154, 19, 4, 98, 17, 140, 144, 72, 80, 18, 2, 0, 0, 0, 0, 0, 0,
	  0, 16, 174, 209, 39, 243, 243, 240},
	 {234, 170, 156, 0, 148, 192, 145, 16, 4, 88, 8, 88, 0, 128, 0, 0, 0, 0, 0,
	  0, 13, 160, 7, 95, 243, 243, 240},
	 {0, 0, 16, 129, 4, 0, 1, 0, 0, 65, 0, 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 212,
	  198, 159, 243, 243, 240},
	 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 16, 0, 0, 11, 170,
	  208, 139, 243, 243, 240},
	 {234, 170, 144, 0, 4, 0, 1, 0, 0, 64, 0, 58, 170, 0, 0, 0, 0, 0, 0, 64, 29,
	  128, 144, 55, 243, 243, 240},
	 {0, 0, 16, 0, 14, 0, 41, 208, 13, 96, 2, 56, 0, 160, 0, 0, 0, 0, 0, 0, 1,
	  30, 133, 111, 243, 243, 240},
	 {0, 64, 0, 60, 0, 41, 0, 16, 16, 15, 15, 10, 74, 64, 0, 0, 0, 0, 0, 0, 27,
	  103, 2, 215, 243, 243, 240},
	 {0, 61, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 64, 0, 0, 0, 0, 0, 0, 0, 190, 1,
	  139, 243, 243, 240},
	 {234, 170, 147, 36, 228, 4, 5, 6, 8, 200, 0, 16, 0, 16, 0, 0, 0, 0, 0, 0,
	  39, 168, 108, 47, 243, 243, 240},
	 {234, 170, 208, 0, 4, 0, 1, 0, 0, 64, 0, 48, 0, 0, 0, 0, 0, 0, 0, 0, 19,
	  53, 171, 39, 243, 243, 240},
	 {234, 170, 144, 0, 4, 48, 65, 4, 16, 64, 32, 16, 0, 0, 0, 0, 0, 0, 0, 0,
	  13, 162, 81, 11, 243, 243, 240},
	 {192, 128, 131, 2, 64, 0, 0, 0, 0, 4, 1, 18, 2, 0, 0, 0, 0, 0, 0, 0, 21,
	  130, 224, 123, 243, 243, 240},
	 {234, 170, 144, 0, 0, 202, 64, 16, 16, 64, 0, 16, 1, 0, 0, 0, 0, 0, 0, 0,
	  31, 66, 212, 7, 243, 243, 240},
	 {0, 0, 0, 0, 0, 0, 0, 42, 160, 0, 0, 0, 0, 0, 0, 0, 16, 32, 0, 0, 13, 163,
	  80, 11, 243, 243, 240},
	 {234, 170, 144, 0, 4, 0, 1, 0, 0, 64, 0, 16, 0, 0, 0, 8, 0, 0, 1, 0, 11,
	  155, 36, 87, 243, 243, 240},
	 {194, 160, 128, 168, 0, 0, 0, 0, 0, 0, 0, 16, 168, 0, 0, 0, 0, 0, 0, 0, 38,
	  45, 18, 203, 243, 243, 240}};
	
	return levels[index];
}



//Reads a board from the first BOARD_SIZE elements of data, an arraw that stores
//a board as raw bytes
Board *loadBoard(unsigned char *data) {
	short i;
	
	Board *board = (Board *)malloc_throw(sizeof(Board));
	char *walls = board->walls;
	Hatch *hatch;
	BitStream *s = newBitStream(data);
	
	char pos;
	char dir;
	
	memset(board, 0, sizeof(Board));
	
	//Load wall data
	for(pos = 0; pos < WIDTH_TIMES_HEIGHT; pos++) {
		if (readBit(s)) {
			walls[pos] += (1 << NORTH);
			walls[advancePos(pos, NORTH)] += (1 << SOUTH);
		}
		
		if (readBit(s)) {
			walls[pos] += (1 << WEST);
			walls[advancePos(pos, WEST)] += (1 << EAST);
		}
	}
	
	//Load pit data
	for(i = 0; i < WIDTH_TIMES_HEIGHT; i++)
		board->pits[i] = readBit(s);
	
	//Load rockets
	board->rocketAPos = readChar(s, 6);
	board->rocketBPos = readChar(s, 6);
	
	//Load hatches
	for(i = 0, hatch = board->hatches; i < MAX_HATCHES; i++, hatch++) {
		pos = readChar(s, 6);
		if (pos >= 0 && pos < WIDTH_TIMES_HEIGHT) {
			dir = readChar(s, 2);
			hatch->pos = pos;
			hatch->displayDir = dir;
			hatch->dir = newDir(dir, walls[pos]);
			(board->numHatches)++;
		}
		else
			break;
	}
	
	freeBitStream(s);
	free(s);
	return board;
}

//Returns whether the specified board is valid
char validateBoard(Board *board) {
	char occupied[WIDTH_TIMES_HEIGHT];
	short i;
	char *walls = board->walls;
	char pos;
	
	memcpy(occupied, board->pits, WIDTH_TIMES_HEIGHT * sizeof(char));
	
	for(i = 0; i < board->numHatches; i++) {
		pos = board->hatches[i].pos;
		if (pos < 0 || pos >= WIDTH_TIMES_HEIGHT ||
				occupied[pos] || walls[pos] == WALLS_ALL)
			return 0;
		occupied[pos] = 1;
	}
	
	pos = board->rocketAPos;
	if (pos < 0 || pos >= WIDTH_TIMES_HEIGHT ||
			occupied[pos] || walls[pos] == WALLS_ALL)
		return 0;
	occupied[pos] = 1;
	
	pos = board->rocketBPos;
	return pos >= 0 && pos < WIDTH_TIMES_HEIGHT &&
		!(occupied[pos]) && walls[pos] != WALLS_ALL;
}

//Loads the built-in board of the specified index.  Assumes that 0 <= index <
//NUM_BUILT_IN_BOARDS.
Board *loadBuiltInBoard(short index) {
	return loadBoard(builtInBoard(index));
}

//Returns a new a game for the specified Board
Game *newGame(Board *board, RNG *rng) {
	short i;
	Game *game = (Game *)malloc_throw(sizeof(Game));
	SpaceData *space;
	char *pits = board->pits;
	char rocketsInverted;
	
	memset(game, 0, sizeof(Game));
	
	game->board = board;
	game->rng = rng;
	
	//Set game->boardData to match board
	for(i = 0, space = game->boardData; i < WIDTH_TIMES_HEIGHT; i++, space++) {
		space->arrowDir = NO_DIRECTION;
		space->type = pits[i] ? SPACE_PIT : SPACE_NOTHING;
	}
	
	rocketsInverted = rngRandom(rng, 2);
	
	game->boardData[board->rocketAPos].type =
		rocketsInverted ? SPACE_ROCKET2 : SPACE_ROCKET1;
	game->boardData[board->rocketBPos].type =
		rocketsInverted ? SPACE_ROCKET1 : SPACE_ROCKET2;
	
	for(i = 0; i < board->numHatches; i++)
		game->boardData[board->hatches[i].pos].type = SPACE_HATCH;
	
	return game;
}

//Frees the contents of the specified Board (but not the pointer to the Board
//itself)
#define freeBoard(x)

//Frees the contents of the specified Game (but not the pointer to the Game
//itself)
void freeGame(Game *game) {
	freeBoard(game->board);
	free(game->board);
}
















#endif
