/*
Created on August 17, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file contains methods pertaining to reading keys in the game.  It provides
functionality for moving the selection cursor and retrieving the direction of an
arrow that the user has indicated to place.  In essence, it takes care of
annoying interrupt handler stuff.  It requires proper usage of the functions
initKeysAndTimer, cleanupKeysAndTimer, switchToGameMode, and switchToMenuMode.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef KEY_H_INCLUDED
#define KEY_H_INCLUDED

#include "game.h"
#include "time.h"

//The number of milliseconds between repeats in movement of the selection cursor
#define REPEAT_DELAY 120
//One more than the number of times AUTO_INT_5 is fired between calls to
//pollKeys
#define KEY_POLL_TICKS 20

//Whether readArrowDir is waiting for an arrow direction.  It is used to signal
//to pollKeys to retrieve and erase an arrow direction.
static volatile char waitingForArrowDir;
//The last arrow direction that was read in pollKeys.  It is used to indicate a
//retrieved direction to readArrowDir.
static volatile char lastArrowDir;

//Keys are represented in binary as wwsseenn, where each pair of bits stores the
//one or two keys used for placing an arrow in the given direction.  The first
//bit for each direction is whether MODE, 'x', 'z', or '(' is pressed on the
//TI-89 and 0 on other calculators.  The second bit is whether '8', '4', '6', or
//'2' is pressed.
//The keys that were depressed the last time readArrowDir was called
static volatile unsigned char prevKeys;
//The keys that have been pressed since the last time readArrowDir was called
static volatile unsigned char pressedKeys;
//The keys that are currently depressed
static volatile unsigned char curKeys;

//Whether ESC was pressed since the last time game mode was switched to
static volatile char pressedEsc0;
//Whether we are waiting for the ESC key to be released before setting
//pressedEsc0 to true
static volatile char escLocked;

//Whether an arrow key was depressed the last time updateSelectedPos was called
static char keyDownLastTime;
//REPEAT_DELAY less than the time when the arrow key press will next repeat,
//moving the selection another time, if keyDownLastTime is true
static long selectionTime;
//Whether we are currently in game mode
static char inGameMode0;

//The original INT_HANDLER for AUTO_INT_1
static INT_HANDLER originalInt1;
//The original INT_HANDLER for AUTO_INT_5
static INT_HANDLER originalInt5;

//Updates and returns the selected position based on the time and the keys that
//are being held down.  Currently assumes that it receives control frequently,
//as it does not check for key presses between calls.  Does not move the cursor
//multiple spaces so that, with careful coding, one can cause behavior that does
//not move the cursor more than once per frame.
char updateSelectedPos(char selectedPos) {
	short dirX = 0;
	short dirY = 0;
	long t;
	char advance;
	
	BEGIN_KEYTEST
	
	if (_keytest_optimized(RR_LEFT))
		dirX--;
	if (_keytest_optimized(RR_RIGHT))
		dirX++;
	if (_keytest_optimized(RR_UP))
		dirY--;
	if (_keytest_optimized(RR_DOWN))
		dirY++;
	
	END_KEYTEST
	
	if (dirX != 0 || dirY != 0) {
		t = time();
		if (keyDownLastTime) {
			advance = selectionTime + REPEAT_DELAY <= t;
			if (advance)
				selectionTime = max(selectionTime + REPEAT_DELAY, t - REPEAT_DELAY);
		}
		else {
			advance = 1;
			keyDownLastTime = 1;
			selectionTime = t;
		}
		
		if (advance) {
			if (dirX)
				selectedPos = advancePos(selectedPos, dirX < 0 ? WEST : EAST);
			if (dirY)
				selectedPos = advancePos(selectedPos, dirY < 0 ? NORTH : SOUTH);
		}
	}
	else
		keyDownLastTime = 0;
	
	return selectedPos;
}

//Checks which of the relevant keys (the ones for dropping arrows) are depressed
static void pollKeys() {
	unsigned char keys = 0;
	unsigned char diff;
	
	//Check each of the relevant keys, and update pressedKeys and curKeys
	BEGIN_KEYTEST
	
	if (_keytest_optimized(RR_4))
		keys += 0x40;
	if (_keytest_optimized(RR_6))
		keys += 0x04;
	if (_keytest_optimized(RR_2))
		keys += 0x10;
	if (_keytest_optimized(RR_8))
		keys += 0x01;
	
	if (TI89) {
		if (_keytest_optimized(RR_X))
			keys += 0x80;
		if (_keytest_optimized(RR_Z))
			keys += 0x08;
		if (_keytest_optimized(RR_PAREN_OPEN))
			keys += 0x20;
		if (_keytest_optimized(RR_MODE))
			keys += 0x02;
	}
	
	if (_keytest_optimized(RR_ESC)) {
		if (!escLocked)
			pressedEsc0 = 1;
	}
	else
		escLocked = 0;
	
	END_KEYTEST
	
	pressedKeys |= keys;
	curKeys = keys;
	
	if (waitingForArrowDir) {
		diff = pressedKeys & ~prevKeys;
		prevKeys = curKeys;
		pressedKeys = 0;
		
		if (diff & 3)
			lastArrowDir = NORTH;
		else if ((diff >> 2) & 3)
			lastArrowDir = EAST;
		else if ((diff >> 4) & 3)
			lastArrowDir = SOUTH;
		else if (diff >> 6)
			lastArrowDir = WEST;
		else
			lastArrowDir = NO_DIRECTION;
		
		waitingForArrowDir = 0;
	}
}

//Function used to periodically poll key data.  Must be placed in AUTO_INT_5.
inline void maybePollKeys() {
	static const unsigned char start = 257 - KEY_POLL_TICKS;
	static unsigned char keyPollCount = start;
	
	if (!(++keyPollCount) || waitingForArrowDir) {
		pollKeys();
		keyPollCount = start;
	}
}

//Returns the direction corresponding to the arrow the user has indicated, or
//NO_DIRECTION if the user has indicated no arrow direction.  The call erases
//this direction from memory.
char readArrowDir() {
	//Tell maybePollKeys that we're waiting for an arrow direction
	waitingForArrowDir = 1;
	//Wait until it is retrieved
	while(waitingForArrowDir);
	//Return the direction
	return lastArrowDir;
}

//Returns whether the user pressed escape since the last call to pressedEsc.
char pressedEsc() {
	char result = pressedEsc0;
	pressedEsc0 = 0;
	return result;
}

//Returns and removes the next key in the queue, or 0 if there is none.  The
//result is or'd with KB_AUTOREPEAT if it is a repeated keystroke.
unsigned short nextKey() {
	unsigned short key;
	return !OSdequeue(&key, kbd_queue()) ? key : 0;
}

//Flushes the keyboard queue
void flushKeyQueue() {
	while (nextKey());
}

DEFINE_INT_HANDLER(GameModeHandler) {
	if (timerTicks < MAXIMUM_TIMER_TICKS)
		timerTicks++;
	maybePollKeys();
}

DEFINE_INT_HANDLER(MenuModeHandler) {
	if (timerTicks < MAXIMUM_TIMER_TICKS)
		timerTicks++;
	ExecuteHandler(originalInt5);
}

//Switches to "game mode".  In game mode, updateSelectedPos and readArrowDir may
//be called, but nextKey and flushKeyQueue may not.  Resets the status of all of
//the keys used in game mode.
void switchToGameMode() {
	prevKeys = 0;
	pressedKeys = 0;
	pressedEsc0 = 0;
	escLocked = _keytest(RR_ESC);
	inGameMode0 = 1;
	SetIntVec(AUTO_INT_1, DUMMY_HANDLER);
	SetIntVec(AUTO_INT_5, GameModeHandler);
}

//Switches to "menu mode".  In menu mode, nextKey and flushKeyQueue may be
//called, but updateSelectedPos and readArrowDir may not.
void switchToMenuMode() {
	SetIntVec(AUTO_INT_1, originalInt1);
	SetIntVec(AUTO_INT_5, MenuModeHandler);
	inGameMode0 = 0;
}

//Whether we are currently in game mode, as opposed to menu mode.  In menu mode,
//nextKey and flushKeyQueue may be called, while in game mode updateSelectedPos
//and readArrowDir may be called.
char inGameMode() {
	return inGameMode0;
}

//Must be called before using any other key functions provided by this file or
//any functions in time.h.  Initializes the game to be in menu mode.
//Automatically calls initTimer.
void initKeysAndTimer() {
	keyDownLastTime = 0;
	inGameMode0 = 0;
	initTimer();
	originalInt1 = GetIntVec(AUTO_INT_1);
	originalInt5 = GetIntVec(AUTO_INT_5);
	switchToMenuMode();
}

//Must be called at the end of the program, after calling initKeysAndTime.
//Automatically calls cleanupTimer.
void cleanupKeysAndTimer() {
	SetIntVec(AUTO_INT_1, originalInt1);
	SetIntVec(AUTO_INT_5, originalInt5);
	cleanupTimer();
}













#endif
