/*
Created on August 9, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file contains the higher-level code regarding executing a linked game.
Everything regarding the state of the game is computed separately on each
calculator, and only arrow data and special messages like a "pause" message are
transmitted between the two.  Arrow messages indicating where the player wants
to place arrows are sent every arrowCycleTime milliseconds.  Arrows are applied
at the message cycle after which they were sent.  Thus, there is a delay of 1 to
2 times arrowCycleTime between when the player presses a key to drop an arrow
and when the arrow is actually dropped.  arrowCycleTime ought to be small enough
to allow for a short delay, but not so small that sending, receiving, and parsing messages slows down the game.

All data received through the link are treated with suspicion.  The program
rejects invalid data that might crash the calculator.

To keep the games identical, a random number generator seeded with the same
value is used on each calculator, the usage of which is carefully controlled.
Thus, seemingly innocent changes to some algorithms may result in
incompatibility with alternative versions of the game.

The rules for executing the game are as follows, with uses of the random number
generator indicated:

- When at one arrow cycle, when there is a conflict between the two players in
arrow placement, it is resolved randomly.  More precisely, pending arrows are
dropped as follows: go through each of player 1's arrows in order.  If the arrow
is at the same position as a pending arrow for player 2, call rngRandom(rng, 2).
If the result is 1, try to drop the arrow for player 1, otherwise, try to drop
it for player 2.  Note that if there is already an arrow at the location (from
some previous cycle), the placement will fail.  Then, drop all arrows for player
2 for which there was not a conflict.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef LINK_MATCH_H_INCLUDED
#define LINK_MATCH_H_INCLUDED

#include "key.h"
#include "menu.h"
#include "messagehandler.h"
#include "rng.h"
#include "time.h"
#include "timedgame.h"

//A (signed) long that uniquely defines the compatibility of this version of the
//game with regard to the link data that are sent and expected.  This includes
//the algorithm used to execute the game.  So if, for example, a new special
//event were added, the link version number would have to be changed.  It is
//possible that a future version of the game will have a new link version,
//but be backwards compatible with an older link version number.  For this
//reason, before each game, both calculators exchange their preferred link
//version numbers, followed by whether they are compatible with the other's
//preferred link version number.  A calculator X plays using the other
//calculator Y's link version number if it can, and either Y can't play using
//X's link version number or Y's link version number is greater.
/*
CalcChu version(s):		Link version number:
1.00 to current			1000
*/
#define LINK_VERSION_NUMBER 1000l
//The number of bytes required to send link match data
#define LINK_MATCH_DATA_SIZE 19
//The maximum number of pending arrows (per cycle)
#define MAX_PENDING_ARROWS 2
//The space that is initially selected
//WIDTH * (HEIGHT / 2) + WIDTH / 2
#define SELECTED_POS_START 31

//Types of messages that can be transmitted to the other calculator.  All of
//these values must be nonnegative and less than 50, as per the constraints
//placed by messagehandler.h on messages.
enum MessageType
	{MESSAGE_START_DATA = 0,
	 MESSAGE_LINK_VERSION_COMPATIBLE,
	 MESSAGE_LINK_VERSION_INCOMPATIBLE,
	 MESSAGE_MATCH_SETUP,
	 MESSAGE_BOARD_DATA,
	 MESSAGE_ARROW_DATA,
	 MESSAGE_PAUSE,
	 MESSAGE_QUIT,
	 MESSAGE_RESUME,
	 MESSAGE_READY,
	 MESSAGE_GOT_IT};

//Return values for updating a game
enum UpdateGameResult {UGR_GAME_NOT_OVER, UGR_GAME_OVER, UGR_QUIT};

//Stores a set of arrows pending placement at some arrow cycle
//14 bytes
typedef struct {
	//The arrows in this
	Arrow arrows[MAX_PENDING_ARROWS];
	//The number of arrows in this
	short numArrows;
} PendingArrows;

//Data used by a linked game
//3901 bytes, plus the size of stored messages in mh
typedef struct {
	//The TimedGame used for this
	TimedGame *tg;
	//The current position of the selection
	char selectedPos;
	//Whether the selection changed since the last frame that was drawn.  This
	//is used to allow for behavior where the cursor moves at most one space per
	//frame.
	char selectedPosChanged;
	//Whether this player is player 1
	char isPlayer1;
	//A buffer of size BOTTOM_BAR_BUFFER_SIZE bytes representing the bottom of
	//the screen storing the match score drawing data
	void *matchDrawBuffer;
	//The random number generator for this
	RNG *rng;
	//The MessageHandler used by this
	MessageHandler *mh;
	//The pending arrows for this player from the last message cycle, if there
	//was a last cycle
	PendingArrows myArrowsLastCycle;
	//The pending arrows for this player for this message cycle.  Arrows may yet
	//accumulate to myArrowsThisCycle.
	PendingArrows myArrowsThisCycle;
	//The pending arrows for the opponent from the last message cycle
	PendingArrows *hisArrowsLastCycle;
	//The pending arrows for the opponent for this message cycle
	PendingArrows *hisArrowsThisCycle;
	//The opposite of whether any arrow cycles have been executed
	char firstArrowCycle;
	//The amount of game time in which to send an arrow message
	long nextArrowMessageTime;
	//The amount of time between consecutive sends of pending arrow data.  Twice
	//this number gives the maximum delay time between pressing the key to drop
	//an arrow and the arrow's actually being dropped.
	long arrowCycleTime;
} LinkGameData;

//Frees the contents of the specified LinkGameData (but not the pointer to the
//LinkGameData itself)
static void freeLinkGameData(LinkGameData *lgd) {
	if (lgd->hisArrowsLastCycle != NULL) {
		free(lgd->hisArrowsLastCycle);
		if (lgd->hisArrowsThisCycle != NULL)
			free(lgd->hisArrowsThisCycle);
	}
	//else, assertion: lgd->hisArrowsThisCycle == NULL
	
	freeTimedGame(lgd->tg);
	free(lgd->tg);
}

//Returns a new LinkGameData for the specified Game
static LinkGameData *newLinkGameData(Game *game,
									 long duration,
									 short speed100,
									 long arrowCycleTime,
									 char isPlayer1,
									 void *matchDrawBuffer,
									 RNG *rng,
									 MessageHandler *mh) {
	LinkGameData *lgd = (LinkGameData *)malloc_throw(sizeof(LinkGameData));
	memset(lgd, 0, sizeof(LinkGameData));
	lgd->isPlayer1 = isPlayer1;
	lgd->tg = newTimedGame(game, duration, speed100, isPlayer1, rng);
	lgd->selectedPos = SELECTED_POS_START;
	lgd->matchDrawBuffer = matchDrawBuffer;
	lgd->rng = rng;
	lgd->mh = mh;
	lgd->nextArrowMessageTime = arrowCycleTime;
	lgd->arrowCycleTime = arrowCycleTime;
	lgd->firstArrowCycle = 1;
	return lgd;
}

//Drops an arrow at the specified position, unless there is already an arrow
//there
static void tryToDropArrow(TimedGame *tg, char isPlayer1, char pos, char dir) {
	if (tg->game->boardData[pos].arrowDir == NO_DIRECTION)
		dropArrow(tg, isPlayer1, pos, dir);
}

//Adds an arrow to the specified PendingArrows.  Checks for the validity of its
//position (i.e. ensures it is not on a pit or on a rocket), and checks whether
//it is in the same position of an arrow in arrows already.
static void addArrow(Game *game, PendingArrows *arrows, char pos, char dir) {
	short i;
	Arrow *arrow;
	
	if (arrows->numArrows < MAX_PENDING_ARROWS &&
			game->boardData[pos].type == SPACE_NOTHING) {
		//Check for arrows in arrows that are at pos
		for(i = 0; i < arrows->numArrows; i++) {
			if (arrows->arrows[i].pos == pos)
				return;
		}
		
		arrow = arrows->arrows + (arrows->numArrows)++;
		arrow->pos = pos;
		arrow->dir = dir;
		arrow->isSmall = 0;
	}
}

//Sends the specified one-byte message.  Returns whether it was successful.
static char sendSimpleMessage(MessageHandler *mh, char type) {
	char *message = (char *)malloc_throw(1);
	message[0] = type;
	return handlerSendMessage(mh, message, 1);
}

//Sends a simple message of the type MESSAGE_GOT_IT.  Returns whether it was
//successful.
static char sendGotIt(MessageHandler *mh) {
	return sendSimpleMessage(mh, MESSAGE_GOT_IT);
}

//Draws a message indicating to wait for the opponent to resume, and sleeps for
//a short time
static void drawWaitForResume() {
	drawMessageAndSleep("Waiting for\nopponent to resume`Press ESC or ON to exit");
}

//Indicates that the opponent has quit
static void opponentQuit() {
	drawOpponentQuit();
	idle();
	while (nextKey() != KEY_ESC);
}

//Waits for the opponent to resume the game.  Returns whether it was successful.
/*
When a game is paused, the sent and received messages are of one of the
following formats:

arrow   arrow       arrow   arrow
pause   arrow       pause   got it
resume  got it  OR  resume  got it
arrow   got it      arrow   arrow
arrow   arrow
*/
static char waitForResume(MessageHandler *mh) {
	unsigned short key;
	ReceiveMessageResult *rmr;
	char c;
	char quit;
	
	drawWaitForResume();
	
	while ((rmr = handlerReceiveMessage(mh))->message == NULL) {
		quit = rmr->quit;
		free(rmr);
		if (quit)
			return 0;
		
		while ((key = nextKey())) {
			if (key == KEY_ESC) {
				sendSimpleMessage(mh, MESSAGE_QUIT);
				return 0;
			}
		}
	}
	
	c = *((char *)(rmr->message->message));
	freeMessage(rmr->message);
	free(rmr);
	if (c != MESSAGE_QUIT)
		return sendGotIt(mh) && sendGotIt(mh);
	else {
		opponentQuit();
		return 0;
	}
}

//Parses the specified message.  Returns whether it was successful.
static char parseMessage(LinkGameData *lgd, Message *message) {
	short i;
	char *data;
	PendingArrows *arrows;
	char pos;
	char dir;
	Game *game = lgd->tg->game;
	
	data = (char *)(message->message);
	switch (data[0]) {
		case MESSAGE_ARROW_DATA:
			arrows = (PendingArrows *)malloc_throw(sizeof(PendingArrows));
			arrows->numArrows = 0;
			
			for(i = 2; i < message->size; i += 2) {
				pos = data[i - 1];
				dir = data[i];
				//This check is a final safeguard against the calculator's
				//crashing and shouldn't occur unless bogus data were sent
				if (pos >= 0 && pos < WIDTH_TIMES_HEIGHT &&
						dir >= 0 && dir <= 3 &&
						game->boardData[pos].type == SPACE_NOTHING)
					addArrow(game, arrows, pos, dir);
			}
			
			//assertion: lgd->hisArrowsThisCycle == NULL
			if (lgd->hisArrowsLastCycle == NULL)
				lgd->hisArrowsLastCycle = arrows;
			else
				lgd->hisArrowsThisCycle = arrows;
			
			//no break
		default:
			return 1;
		case MESSAGE_PAUSE:
			return waitForResume(lgd->mh);
		case MESSAGE_QUIT:
			opponentQuit();
			return 0;
	}
}

//Sends the specified PendingArrows.  Returns whether it was successful.
static char sendPendingArrows(MessageHandler *mh, PendingArrows *arrows) {
	short i;
	short size = arrows->numArrows * 2 + 1;
	char *data = (char *)malloc_throw(size);
	Arrow *arrow;
	
	data[0] = MESSAGE_ARROW_DATA;
	for(i = 1, arrow = arrows->arrows; i < size; i += 2, arrow++) {
		data[i] = arrow->pos;
		data[i + 1] = arrow->dir;
	}
	return handlerSendMessage(mh, data, size);
}

//Ensures that lgd->hisArrowsLastCycle != NULL by waiting for an arrow message
//if appropriate.  If wait is true, it assumes that the game is in progress, so
//it waits to try to synchronize the game if this calculator is too far ahead.
static char receiveNewArrows(LinkGameData *lgd, char wait) {
	MessageAndTime *mat;
	Message *message;
	long t = 0;
	char success;
	
	while (lgd->hisArrowsLastCycle == NULL) {
		mat = handlerWaitForMessage(lgd->mh, t);
		message = mat->message;
		t = mat->time;
		free(mat);
		if (message == NULL)
			return 0;
		success = parseMessage(lgd, message);
		freeMessage(message);
		if (!success)
			return 0;
	}
	
	//If we had to wait for the message, then we are ahead of the other player.
	//Let's delay a while so that we are about one half of a cycle ahead of the
	//other player.  Note that without the delay, the game will play fine, but
	//the calculator that is ahead will start skipping frames.
	if (t > 0 && wait)
		delayTimedGame(lgd->tg, t + lgd->arrowCycleTime / 2);
	return 1;
}

/*
//Retrieves and removes the next pending arrows for the opponent
//This function is manually inlined the one place it would have been used
PendingArrows *dequeueHisPendingArrows(LinkGameData *lgd) {
	PendingArrows *arrows = lgd->hisArrowsLastCycle;
	lgd->hisArrowsLastCycle = lgd->hisArrowsThisCycle;
	lgd->hisArrowsThisCycle = NULL;
	return arrows;
}
*/

//Handles the beginning of a new arrow cycle.  To do this, it applies the
//appropriate pending arrows, receives pending arrows for the opponent (if they
//have not been already), and sends the accumulated pending arrows.  Returns
//whether it was successful.
static char updateArrows(LinkGameData *lgd) {
	short i;
	Arrow *arrow;
	PendingArrows *p1PendingArrows;
	PendingArrows *p2PendingArrows;
	char p1Arrows[WIDTH_TIMES_HEIGHT];
	char p2ArrowDirs[WIDTH_TIMES_HEIGHT];
	char pos;
	
	memset(p1Arrows, 0, WIDTH_TIMES_HEIGHT);
	memset(p2ArrowDirs, NO_DIRECTION, WIDTH_TIMES_HEIGHT);
	
	//Send and receive new pending arrows
	if ((!(lgd->firstArrowCycle) && !receiveNewArrows(lgd, 1)) ||
			  !sendPendingArrows(lgd->mh, &(lgd->myArrowsThisCycle)))
		return 0;
	
	if (!(lgd->firstArrowCycle)) {
		//Drop pending arrows
		if (lgd->isPlayer1) {
			p1PendingArrows = &(lgd->myArrowsLastCycle);
			p2PendingArrows = lgd->hisArrowsLastCycle;
		}
		else {
			p1PendingArrows = lgd->hisArrowsLastCycle;
			p2PendingArrows = &(lgd->myArrowsLastCycle);
		}
		
		for(i = 0, arrow = p2PendingArrows->arrows;
				i < p2PendingArrows->numArrows;
				i++, arrow++)
			p2ArrowDirs[arrow->pos] = arrow->dir;
		
		for(i = 0, arrow = p1PendingArrows->arrows;
				i < p1PendingArrows->numArrows;
				i++, arrow++) {
			pos = arrow->pos;
			//When there is a conflict, select a random winner
			if (p2ArrowDirs[pos] == NO_DIRECTION || rngRandom(lgd->rng, 2))
				tryToDropArrow(lgd->tg, 1, pos, arrow->dir);
			else
				tryToDropArrow(lgd->tg, 0, pos, p2ArrowDirs[pos]);
			p1Arrows[pos] = 1;
		}
		
		for(i = 0, arrow = p2PendingArrows->arrows;
				i < p2PendingArrows->numArrows;
				i++, arrow++) {
			if (!p1Arrows[i])
				tryToDropArrow(lgd->tg, 0, arrow->pos, arrow->dir);
		}
		
		free(lgd->hisArrowsLastCycle);
		lgd->hisArrowsLastCycle = lgd->hisArrowsThisCycle;
		lgd->hisArrowsThisCycle = NULL;
	}
	else
		lgd->firstArrowCycle = 0;
	
	memcpy(&(lgd->myArrowsLastCycle), &(lgd->myArrowsThisCycle),
		   sizeof(PendingArrows));
	lgd->myArrowsThisCycle.numArrows = 0;
	return 1;
}

//Waits for and discards the next pending message.  Returns whether it was
//successful.  Normally used to discard "got it" messages.
static char trashMessage(MessageHandler *mh) {
	MessageAndTime *mat = handlerWaitForMessage(mh, 0);
	Message *message = mat->message;
	free(mat);
	
	if (message != NULL) {
		freeMessage(message);
		return 1;
	}
	else
		return 0;
}

//The data type passed to parseAction
typedef struct {
	//The LinkGameData used to parse messages
	LinkGameData *lgd;
	//When parseAction returns, this indicates whether a message of type
	//MESSAGE_QUIT was parsed in the process
	char parsedQuitMessage;
} ParseActionData;

//Action for use while presenting a menu that parses incoming messages.  This is
//meant to be used to check for a message of type MESSAGE_QUIT.  data is taken
//to be of type ParseActionData.
static char parseAction(void *data) {
	char result;
	ParseActionData *pad = (ParseActionData *)data;
	Message *message;
	ReceiveMessageResult *rmr = handlerReceiveMessage(pad->lgd->mh);
	
	if (rmr->quit) {
		free(rmr);
		pad->parsedQuitMessage = 0;
		return 0;
	}
	
	message = rmr->message;
	free(rmr);
	
	if (message != NULL) {
		result = parseMessage(pad->lgd, message);
		freeMessage(message);
		pad->parsedQuitMessage = !result;
		return result;
	}
	else {
		pad->parsedQuitMessage = 0;
		return 1;
	}
}

//Pauses the game.  Returns false if and only if the game should be quit, due to
//a link transmission error or the user's selecting to quit.
/*
When a game is paused, the sent and received messages are of one of the
following formats:

arrow   arrow       arrow   arrow
pause   arrow       pause   got it
resume  got it  OR  resume  got it
arrow   got it      arrow   arrow
arrow   arrow
*/
static char pauseGame(LinkGameData *lgd) {
	ParseActionData pad;
	Menu *curMenu;
	Menu *exitMenu;
	Menu *confirmMenu;
	char quit;
	short s;
	short *results;
	char first = 1;
	
	pad.lgd = lgd;
	
	if (!receiveNewArrows(lgd, 0) || !sendSimpleMessage(lgd->mh, MESSAGE_PAUSE))
		return 0;
	
	exitMenu = newMenu(1);
	addMenuItem(exitMenu, newMenuItem("Game paused", 0, F_8x10), 8);
	addMenuItem(exitMenu, newMenuItem("Resume", 1, F_6x8), 1);
	addMenuItem(exitMenu, newMenuItem("Quit", 1, F_6x8), 0);
	
	confirmMenu = newMenu(1);
	addMenuItem(confirmMenu, newMenuItem("Really quit?", 0, F_8x10), 8);
	addMenuItem(confirmMenu, newMenuItem("No", 1, F_6x8), 1);
	addMenuItem(confirmMenu, newMenuItem("Yes", 1, F_6x8), 0);
	
	curMenu = exitMenu;
	
	while (1) {
		first = 0;
		
		results =
			presentMenuWithAction(curMenu, first ? 1000 : 0, parseAction, &pad);
		if (results != NULL) {
			s = results[0];
			free(results);
		}
		else if (!pad.parsedQuitMessage)
			s = -1;
		else {
			quit = 1;
			break;
		}
		
		if (curMenu == exitMenu) {
			if (s <= 0) {
				quit = 0;
				break;
			}
			else
				curMenu = confirmMenu;
		}
		else if (s <= 0)
			curMenu = exitMenu;
		else {
			quit = 1;
			break;
		}
	}
	
	freeMenu(exitMenu);
	free(exitMenu);
	freeMenu(confirmMenu);
	free(confirmMenu);
	
	return sendSimpleMessage(lgd->mh, quit ? MESSAGE_QUIT : MESSAGE_RESUME) &&
		!quit && receiveNewArrows(lgd, 0);
}

//Handles all of the keys in the queue.  Returns whether it was successful.
static char handleKeys(LinkGameData *lgd) {
	char selectedPos;
	char dir;
	char allowNormalInput = allowInput(lgd->tg);
	
	if (!inGameMode())
		switchToGameMode();
	
	if (allowNormalInput && !(lgd->selectedPosChanged)) {
		selectedPos = updateSelectedPos(lgd->selectedPos);
		if (lgd->selectedPos != selectedPos) {
			lgd->selectedPosChanged = 1;
			lgd->selectedPos = selectedPos;
		}
	}
	
	if (pressedEsc())
		return pauseGame(lgd);
	
	dir = readArrowDir();
	if (dir != NO_DIRECTION && allowNormalInput)
		addArrow(lgd->tg->game,
				 &(lgd->myArrowsThisCycle),
				 lgd->selectedPos,
				 dir);
	return 1;
}

//Replaces the first length bytes of dest with a bitwise or between dest and
//src.  This would probably be faster if it operated using longs rather than
//unsigned chars, although there would be memory alignment issues.
static void orBuffers(void *dest, void *src, short length) {
	short i;
	for(i = 0; i < length; i++)
		*((unsigned char *)dest++) |= *((unsigned char *)src++);
}

//Draws the specified linked game
static void drawLinkGame(LinkGameData *lgd) {
	char buffer[EXPANDED_DRAW_SIZE];
	char *buffer2 = buffer + EXPANDED_DRAW_OFFSET;
	
	drawTimedGameBackground(buffer2, lgd->tg);
	orBuffers(buffer2 + BOTTOM_BAR_BUFFER_OFFSET,
			  lgd->matchDrawBuffer,
			  BOTTOM_BAR_BUFFER_SIZE);
	drawSelection(buffer2, lgd->selectedPos);
	drawTimedGameForeground(buffer2, lgd->tg);
	draw89Size(buffer2);
	lgd->selectedPosChanged = 0;
}

//Updates the specified linked game by one or zero frames.  Returns
//UGR_GAME_NOT_OVER, UGR_GAME_OVER, or UGR_QUIT.  Profiling suggests that in bad
//cases (i.e. lots of critters), about two-thirds of the time is spent in
//drawLinkGame, and about half of this time is spent in drawCritters.
static char updateLinkGame(LinkGameData *lgd) {
	TimedGame *tg = lgd->tg;
	
	//Note that arrow messages are sent every lgd->arrowCycleTime real time, not
	//game time (which excludes time spent displaying messages), because this
	//simplifies handling pausing while messages are being displayed and more
	//easily allows for near-instantaneous recognition of such an event by the
	//other calculator when a message is being displayed
	lgd->nextArrowMessageTime -= updateTimedGame(tg, lgd->nextArrowMessageTime);
	if (lgd->nextArrowMessageTime <= 0) {
		if (!updateArrows(lgd))
			return UGR_QUIT;
		lgd->nextArrowMessageTime =
			lgd->arrowCycleTime - updateTimedGame(tg, lgd->arrowCycleTime);
	}
	
	
	
	if (gameOver(tg)) {
		//Send a garbagy set of pending arrows.  Doing so is necessary for
		//pausing to work at the end of the game.
		if (receiveNewArrows(lgd, 0) &&
				sendPendingArrows(lgd->mh, &(lgd->myArrowsThisCycle))) {
			//Remove one set of pending arrows
			if (lgd->hisArrowsLastCycle != NULL)
				free(lgd->hisArrowsLastCycle);
			lgd->hisArrowsLastCycle = lgd->hisArrowsThisCycle;
			lgd->hisArrowsThisCycle = NULL;
			
			return receiveNewArrows(lgd, 0) ? UGR_GAME_OVER : UGR_QUIT;
		}
		else
			return UGR_QUIT;
	}
	
	if (!handleKeys(lgd))
		return UGR_QUIT;
	
	//Drawing takes up the bulk of the time in this function, about two thirds
	//of the time in the worse cases (i.e. lots of critters)
	drawLinkGame(lgd);
	
	return UGR_GAME_NOT_OVER;
}





//Stores data for a linked match relevant to all players
//30 bytes
typedef struct {
	//The number of games to which this match is being played
	short firstTo;
	//The score handicap of the game.  A negative handicap indicates that player
	//1 starts with a number of points equal to its absolute value, while a
	//positive handicap indicates that player 2 starts with that number of
	//points.
	short handicap;
	//The duration of games
	long duration;
	//The number of games player 1 has won
	short score1;
	//The number of games player 2 has won
	short score2;
	//A speed modifier for the game
	short speed100;
	//The seed used to initialize the random number generator for this match.  A
	//lot depends on proper synchronization of random values.
	unsigned long seed;
	//The amount of time between consecutive sends of pending arrow data.  Twice
	//this number gives the maximum delay time between pressing the key to drop
	//an arrow and the arrow's actually being dropped.
	long arrowCycleTime;
} LinkMatchData;

//Frees the contents of the specified LinkMatchData (but not the pointer to the
//LinkMatchData itself)
#define freeLinkMatchData(x)

//Stores data for a linked match relevant to the host
//59 bytes
typedef struct {
	//Whether each board may be played next, so that no board is played multiple
	//times until all boards are exhaused
	char boardAvailable[NUM_BUILT_IN_BOARDS];
	//The number of false elements in boardAvailable
	short numBoardsUnavailable;
	//The index of the last board that was played, or a negative number if there
	//is no such board
	short lastBoard;
	LinkMatchData matchData;
} LinkMatchHostData;

//Menu options
#define FIRST_TO_VALUES ((const char *[]){"1", "2", "3", "4", "5"})
#define DURATION_VALUES \
	((const char *[]){"60", "90", "120", "180", "240", "300"})
const char *SPEED_VALUES[] =
	{"very slow", "slow", "normal", "fast", "very fast"};
#define HANDICAP_VALUES ((const char *[]) \
	{"200 for him", \
	 "100 for him", \
	 "50 for him", \
	 "none", \
	 "50 for you", \
	 "100 for you", \
	 "200 for you"})

#define DURATIONS \
	((const long[6]){60999, 90999, 120999, 180999, 240999, 300999})
#define SPEEDS100 ((const short[5]){60, 80, 100, 135, 170})
#define ARROW_CYCLE_TIMES ((const short[5]){500, 350, 200, 100, 50})
#define HANDICAPS ((const short[7]){200, 100, 50, 0, -50, -100, -200})

//Sets up a linked match by retrieving user input.  Returns NULL if the user
//chose to cancel.
static LinkMatchHostData *setupLinkMatch() {
	LinkMatchHostData *lmhd;
	LinkMatchData *lmd;
	Menu *menu = newMenu(1);
	short *results;
	
	addMenuItem(menu, newMenuItem("Setup Match", 0, F_8x10), 3);
	addMenuItem(menu, newMenuItemWithValues("First to:",
											FIRST_TO_VALUES, 5, 2, F_6x8), 0);
	addMenuItem(menu, newMenuItemWithValues("Duration:",
											DURATION_VALUES, 6, 2, F_6x8), 0);
	addMenuItem(menu, newMenuItemWithValues("Speed:",
											SPEED_VALUES, 5, 2, F_6x8), 0);
	addMenuItem(menu, newMenuItemWithValues("Handicap:",
											HANDICAP_VALUES, 7, 3, F_6x8), 0);
	addMenuItem(menu, newMenuItemWithValues("Link speed:",
											SPEED_VALUES, 5, 3, F_6x8), 0);
	addMenuItem(menu, newMenuItem("Decrease link speed if game runs too slowly",
								  0, F_4x6), 3);
	addMenuItem(menu, newMenuItem("Cancel", 1, F_6x8), 0);
	addMenuItem(menu, newMenuItem("Done", 1, F_6x8), 0);
	results = presentMenu(menu, 0);
	freeMenu(menu);
	free(menu);
	
	if (results == NULL)
		return NULL;
	if (results[0] == 0) {
		free(results);
		return NULL;
	}
	
	lmhd = (LinkMatchHostData *)malloc_throw(sizeof(LinkMatchHostData));
	lmd = &(lmhd->matchData);
	memset(lmhd->boardAvailable, 1, NUM_BUILT_IN_BOARDS);
	lmhd->lastBoard = -1;
	lmhd->numBoardsUnavailable = 0;
	lmd->score1 = 0;
	lmd->score2 = 0;
	
	//Certain values in the unsigned long range are impossible, but this isn't
	//of concern
	lmd->seed = (((unsigned long)rand()) * 32768l) + ((unsigned long)rand());
	lmd->firstTo = results[1] + 1;
	lmd->duration = DURATIONS[results[2]];
	lmd->speed100 = SPEEDS100[results[3]];
	lmd->handicap = HANDICAPS[results[4]];
	lmd->arrowCycleTime = ARROW_CYCLE_TIMES[results[5]];
	
	free(results);
	return lmhd;
}

//Frees the contents of the specified LinkMatchHostData (but not the pointer to
//the LinkMatchHostData itself)
static void freeLinkMatchHostData(LinkMatchHostData *lmhd) {
	freeLinkMatchData(&(lmhd->matchData));
}

//Returns an index for the next board to play, given which boards are available.
//The algorithm uses shuffling, so that no board is replayed until all have been
//played and no board is played twice in a row.  Updates boardAvailable,
//lastBoard, and numBoardsUnavailable as appropriate.  numBoardsUnavailable is a
//pointer to the number of false elements in boardAvailable.
static short boardIndex(char *boardAvailable,
						short *numBoardsUnavailable,
						short *lastBoard) {
	short i;
	
	while (!boardAvailable[i = random(NUM_BUILT_IN_BOARDS)] || i == *lastBoard);
	boardAvailable[i] = 0;
	*lastBoard = i;
	
	if (++(*numBoardsUnavailable) == NUM_BUILT_IN_BOARDS) {
		memset(boardAvailable, 1, NUM_BUILT_IN_BOARDS);
		*numBoardsUnavailable = 0;
	}
	return i;
}

//Transmits the specified LinkMatchData.  Returns whether it was successful.
static char sendLinkMatchData(MessageHandler *mh, LinkMatchData *lmd) {
	char *bytes = (char *)malloc_throw(LINK_MATCH_DATA_SIZE);
	
	bytes[0] = MESSAGE_MATCH_SETUP;
	memcpy(bytes + 1, &(lmd->firstTo), 2);
	memcpy(bytes + 3, &(lmd->handicap), 2);
	memcpy(bytes + 5, &(lmd->duration), 4);
	memcpy(bytes + 9, &(lmd->speed100), 2);
	memcpy(bytes + 11, &(lmd->seed), 4);
	memcpy(bytes + 15, &(lmd->arrowCycleTime), 4);
	return handlerSendMessage(mh, bytes, LINK_MATCH_DATA_SIZE) &&
		trashMessage(mh);
}

//Draws a message indicating that the user is waiting for someone to join the
//match
static void drawWaitingForSomeoneToJoin() {
	drawMessage("Waiting for\nsomeone to join`Press ESC or ON to cancel");
}

//Draws a message indicating that the link version of the game on this
//calculator does not match that of the other, and sleeps for a short time
static void drawWrongVersion() {
	drawMessageAndSleep("Error:|The other calculator is\nusing an incompatible"
						"\nversion of the game`Press ESC to exit");
}

//Draws a message indicating that both players are attempting to join a game,
//and sleeps for a short time
static void drawBothJoining() {
	drawMessageAndSleep("Error:|You can't both join a\ngame - one player must\n"
						"be the host`Press ESC to exit");
}

//Draws a message indicating that both players are attempting to host a game,
//and sleeps for a short time
static void drawBothHosting() {
	drawMessageAndSleep("Error:|You can't both host a game`Press ESC to exit");
}

//Draws a message indicating that the user is joining a match
static void drawJoining() {
	drawMessage("Joining`Press ESC or ON to cancel");
}

//Exchanges the link version numbers and whether this is trying to host the
//game.  Returns whether to continue, i.e. whether the values sent by the other
//calculator indicate a compatible game.
static char exchangeStartingData(MessageHandler *mh, char isHost) {
	MessageAndTime *mat;
	Message *message;
	char *buffer = (char *)malloc_throw(6);
	long l;
	long t;
	char compatible;
	char dontQuit;
	char c;
	
	buffer[0] = MESSAGE_START_DATA;
	l = LINK_VERSION_NUMBER;
	memcpy(buffer + 1, &l, 4);
	buffer[5] = isHost;
	handlerSendMessage(mh, buffer, 6);
	
	t = 0;
	while ((mat = handlerWaitForMessage(mh, t))->message != NULL) {
		t = mat->time;
		message = mat->message;
		free(mat);
		
		buffer = (char *)(message->message);
		if (message->size == 6 && buffer[0] == MESSAGE_START_DATA) {
			//Get starting data
			if (!(buffer[5] ^ isHost)) {
				if (isHost)
					drawBothHosting();
				else
					drawBothJoining();
				freeMessage(message);
				idle();
				while (nextKey() != KEY_ESC);
				return 0;
			}
			
			memcpy(&l, buffer + 1, 4);
			freeMessage(message);
			
			//Send a message indicating whether this calculator is capable of
			//playing using the other calculator's preferred link version number
			compatible = (l == LINK_VERSION_NUMBER);
			buffer = (char *)malloc_throw(1);
			buffer[0] = compatible ? MESSAGE_LINK_VERSION_COMPATIBLE :
					MESSAGE_LINK_VERSION_INCOMPATIBLE;
			handlerSendMessage(mh, buffer, 1);
			t = 0;
			
			//See whether the other calculator is capable of playing using this
			//calculator's preferred link version number
			while ((dontQuit =
					(mat = handlerWaitForMessage(mh, t))->message != NULL)) {
				t = mat->time;
				message = mat->message;
				c = ((char *)(message->message))[0];
				freeMessage(message);
				free(mat);
				switch(c) {
					case MESSAGE_LINK_VERSION_COMPATIBLE:
						compatible = 1;
					case MESSAGE_LINK_VERSION_INCOMPATIBLE:
						goto BreakOuter;
				}
			}
BreakOuter:
			if (dontQuit) {
				if (compatible)
					return 1;
				else {
					drawWrongVersion();
					idle();
					while (nextKey() != KEY_ESC);
					return 0;
				}
			}
			else {
				free(mat);
				return 0;
			}
		}
		
		freeMessage(message);
	}
	
	free(mat);
	return 0;
}

//Draws a message indicating that the player should get ready to start and waits
//for one of the players to press ENTER.  Returns whether it was successful.
static char getReady(MessageHandler *mh, char isPlayer1) {
	short key;
	ReceiveMessageResult *rmr;
	
	drawMessage(isPlayer1 ?
				"You will be\nwhite this match`Press ENTER to start" :
				"You will be\nblack this match`Press ENTER to start");
	flushKeyQueue();
	
	while ((rmr = handlerReceiveMessage(mh))->message == NULL) {
		if (rmr->quit) {
			free(rmr);
			return 0;
		}
		
		free(rmr);
		
		while ((key = nextKey())) {
			if (key == KEY_ENTER)
				return sendSimpleMessage(mh, MESSAGE_READY) && trashMessage(mh);
		}
	}
	
	freeMessage(rmr->message);
	free(rmr);
	return sendGotIt(mh);
}

//Attempts to host a new session with the specified linkMatchData.  Returns
//whether it was successful.
static char hostMatchSession(MessageHandler *mh, LinkMatchData *lmd) {
	unsigned short key;
	
	drawWaitingForSomeoneToJoin();
	while (!handlerStartSession(mh)) {
		while ((key = nextKey())) {
			if (key == KEY_ESC) {
				handlerAbortSessionAttempt(mh);
				return 0;
			}
		}
	}
	
	if (exchangeStartingData(mh, 1) &&
		   sendLinkMatchData(mh, lmd) &&
		   getReady(mh, 1))
		return 1;
	else {
		handlerCloseSession(mh);
		return 0;
	}
}

//Executes a game on the specified board.  Returns whether it was successfully
//completed.  Ensures that the board is freed when done.
static char executeGame(Board *board,
						LinkMatchData *lmd,
						char isPlayer1,
						RNG *rng,
						MessageHandler *mh) {
	LinkGameData *lgd;
	char matchDrawBuffer[BOTTOM_BAR_BUFFER_SIZE];
	Game *game = newGame(board, rng);
	char val = UGR_QUIT;
	
	if (lmd->handicap < 0)
		game->score1 = -(lmd->handicap);
	else
		game->score2 = lmd->handicap;
	
	drawMatchScore(matchDrawBuffer, lmd->score1, lmd->score2, lmd->firstTo);
	lgd = newLinkGameData(game,
						  lmd->duration,
						  lmd->speed100,
						  lmd->arrowCycleTime,
						  isPlayer1,
						  matchDrawBuffer,
						  rng,
						  mh);
	
	TRY
	
	//Needed to reset the status of the keys used in game mode
	switchToGameMode();
	
	while ((val = updateLinkGame(lgd)) == UGR_GAME_NOT_OVER);
	
	if (val != UGR_QUIT) {
		if (game->score1 > game->score2)
			(lmd->score1)++;
		else if (game->score2 > game->score1)
			(lmd->score2)++;
	}
	
	FINALLY
	
	//Try to recover at least some memory if it crashes (which is, of course,
	//impossible, because I never write code that has any bugs in it)
	freeGame(game);
	free(game);
	freeLinkGameData(lgd);
	free(lgd);
	
	ENDFINAL
	
	#if 0
	ClrScr();
	printf("%3f %3f %3f %3f %3f", AVG(T1), AVG(T2), AVG(T3), AVG(T4), AVG(T5));
	while(1);
	#endif
	
	return val == UGR_GAME_OVER;
}

//Sends the specified board data.  Returns whether it was successful.  Note that
//entire boards are sent rather than board indices in case new boards are added
//later (or in case of a board editor, etc.).
static char sendBoardData(MessageHandler *mh, unsigned char *boardData) {
	char *message = (char *)malloc_throw(BOARD_SIZE + 1);
	message[0] = MESSAGE_BOARD_DATA;
	memcpy(message + 1, boardData, BOARD_SIZE);
	return handlerSendMessage(mh, message, BOARD_SIZE + 1) && trashMessage(mh);
}

//Returns whether the specified LinkMatchData is valid, ignoring the score1 and
//score2 fields
static char validateLinkMatchData(LinkMatchData *lmd) {
	return lmd->firstTo > 0 && lmd->firstTo <= 5 &&
			abs(lmd->handicap) <= 999 &&
			lmd->duration > 0 && lmd->duration < 1000000 &&
			lmd->speed100 > 0 && lmd->speed100 < 2000 &&
			lmd->arrowCycleTime >= 5 && lmd->arrowCycleTime <= 5000;
}

//Returns a LinkMatchData from the link if one is received and NULL otherwise
static LinkMatchData *receiveLinkMatchData(MessageHandler *mh) {
	LinkMatchData *lmd = NULL;
	MessageAndTime *mat;
	Message *message;
	char *data;
	long t = 0;
	
	while ((mat = handlerWaitForMessage(mh, t))->message != NULL) {
		t = mat->time;
		message = mat->message;
		free(mat);
		
		data = (char *)(message->message);
		if (message->size == LINK_MATCH_DATA_SIZE &&
				data[0] == MESSAGE_MATCH_SETUP) {
			lmd = (LinkMatchData *)malloc_throw(sizeof(LinkMatchData));
			memcpy(&(lmd->firstTo), data + 1, 2);
			memcpy(&(lmd->handicap), data + 3, 2);
			memcpy(&(lmd->duration), data + 5, 4);
			memcpy(&(lmd->speed100), data + 9, 2);
			memcpy(&(lmd->seed), data + 11, 4);
			memcpy(&(lmd->arrowCycleTime), data + 15, 4);
			lmd->score1 = 0;
			lmd->score2 = 0;
			if (validateLinkMatchData(lmd)) {
				freeMessage(message);
				sendGotIt(mh);
				return lmd;
			}
			else {
				freeLinkMatchData(lmd);
				free(lmd);
			}
		}
		
		freeMessage(message);
	}
	
	free(mat);
	return NULL;
}

//Attempts to join a new session.  Returns the match data if it was successful
//and NULL otherwise.
static LinkMatchData *joinMatchSession(MessageHandler *mh) {
	unsigned short key;
	LinkMatchData *lmd;
	
	drawJoining();
	while (!handlerStartSession(mh)) {
		while ((key = nextKey())) {
			if (key == KEY_ESC) {
				handlerAbortSessionAttempt(mh);
				return NULL;
			}
		}
	}
	
	if (exchangeStartingData(mh, 0) &&
			(lmd = receiveLinkMatchData(mh)) != NULL) {
		if (getReady(mh, 0))
			return lmd;
		else {
			freeLinkMatchData(lmd);
			free(lmd);
		}
	}
	
	handlerCloseSession(mh);
	return NULL;
}

//Retrieves board data from the link.  Returns NULL if it was unsuccessful.
static Board *receiveBoardData(MessageHandler *mh) {
	Board *board;
	MessageAndTime *mat;
	Message *message;
	long t = 0;
	
	while ((mat = handlerWaitForMessage(mh, t))->message != NULL) {
		t = mat->time;
		message = mat->message;
		free(mat);
		
		if (message->size == BOARD_SIZE + 1 &&
				*((char *)(message->message)) == MESSAGE_BOARD_DATA) {
			board = loadBoard((unsigned char *)(message->message + 1));
			freeMessage(message);
			
			//Validate the board, to make sure it doesn't crash the calculator
			//or anything
			if (validateBoard(board) && sendGotIt(mh))
				return board;
			else {
				freeBoard(board);
				free(board);
			}
		}
		else
			freeMessage(message);
	}
	
	free(mat);
	return NULL;
}

//Displays the winner of the match
static void showWinner(short score1, short score2) {
	short key;
	
	drawWinner(score1, score2);
	idle();
	while ((key = nextKey()) != KEY_ESC && key != KEY_ENTER);
}

//Hosts and executes a linked match
void hostLinkMatch() {
	MessageHandler *mh = newMessageHandler();
	Board *board;
	RNG *rng;
	LinkMatchData *lmd;
	unsigned char *boardData;
	LinkMatchHostData *lmhd;
	char nooneWon;
	
	TRY
	
	lmhd = setupLinkMatch();
	if (lmhd != NULL) {
		lmd = &(lmhd->matchData);
		if (hostMatchSession(mh, lmd)) {
			rng = newRNG(lmd->seed);
			resetTimer();
			while((nooneWon = (max(lmd->score1, lmd->score2) < lmd->firstTo))) {
				boardData =
					builtInBoard(boardIndex(lmhd->boardAvailable,
											&(lmhd->numBoardsUnavailable),
											&(lmhd->lastBoard)));
				if (!sendBoardData(mh, boardData))
					break;
				
				board = loadBoard(boardData);
				//assertion: validateBoard(board) is true
				if (!executeGame(board, lmd, 1, rng, mh))
					break;
			}
			
			freeRNG(rng);
			free(rng);
			handlerCloseSession(mh);
			
			if (!nooneWon)
				showWinner(lmd->score1, lmd->score2);
		}
		
		freeLinkMatchHostData(lmhd);
		free(lmhd);
	}
	
	FINALLY
	
	//Try to recover at least some memory if it crashes (which is, of course,
	//impossible, because I never write code that has any bugs in it)
	freeMessageHandler(mh);
	free(mh);
	
	ENDFINAL
}

//Joins and executes a linked match
void joinLinkMatch() {
	Board *board;
	RNG *rng;
	MessageHandler *mh = newMessageHandler();
	LinkMatchData *lmd;
	char nooneWon;
	
	TRY
	
	lmd = joinMatchSession(mh);
	if (lmd != NULL) {
		rng = newRNG(lmd->seed);
		resetTimer();
		while((nooneWon = (max(lmd->score1, lmd->score2) < lmd->firstTo)) &&
					 (board = receiveBoardData(mh)) != NULL &&
					 executeGame(board, lmd, 0, rng, mh));
		
		freeRNG(rng);
		free(rng);
		handlerCloseSession(mh);
		
		if (!nooneWon)
			showWinner(lmd->score2, lmd->score1);
		
		freeLinkMatchData(lmd);
		free(lmd);
	}
	
	FINALLY
	
	//Try to recover at least some memory if it crashes (which is, of course,
	//impossible, because I never write code that has any bugs in it)
	freeMessageHandler(mh);
	free(mh);
	
	ENDFINAL
}
















#endif
