/*
Created on August 15, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file provides a system for message sending and receiving that checks
incoming messages for data integrity, defined to be the quality of received data
matching sent data.  Note that this check is fairly expensive with regard to
program size and complexity and transfer rate.  In the case of CalcChu, this
increases the required transfer rate roughly by a factor of 3, as a lot of small
messages are sent in the program.  The cost is believed to be worth the
benefits.

Each message is sent by sending a character indicating its length, then the
contents of the message, then the digest of the message, which chaotically (but
deterministically) reduces the contents of the message to a long.  An SOS signal
may be sent by sending 16 bytes that encode a unique ID for the session with the
other calculator.  Received data may be interpreted in any manner consistent
with receiving any set of sent bytes in the proper order, provided that if they
are all received, all messages are correctly interpreted (unless there is an
ambiguous case, which is very, very unlikely; see below).  Note that this
necessitates comparing the digest of the contents of a potential message with
the four bytes immediately after the message.  Generally, when an SOS signal is
received, the next byte ought to be the beginning of a message.

A message digest is generated by taking some starting value and for each
character in the message, performing some operation on the current digest value
and the character to obtain a new digest value.  The starting digest and session
ID are initialized to particular starting values until a session is established.
A session is started by repeatedly sending an SOS signal followed by a message
containing a fixed random long value.  Once a message is received, the starting
digest is set to be the xor of the two random values, and the session ID is
altered so that the first four bytes contain the lesser of the two values and
the last four contain the greater of the two.

Regarding the consideration that data may validly be interpreted wrongly: there
are three ways this may happen.  First, data that happen to be the same as the
session ID may be sent and interpreted as an SOS signal.  The SOS signal is 16
bytes long, so if random data were sent, there would be a 1 in ~3*10^38 chance
of this for any 16-byte segment of received bytes.  So even if everyone in the
world downloaded and used this software continuously for the next hundred years,
it would be highly unlikely that even one collision would occur.  Even if the
data are highly ordered, the last 8 bytes of the session ID were randomly
generated in advance, and it is unlikely that the order of the data is related
to the "order" in these randomly generated bytes.  This leaves at most a 1 in
~2*10^19 chance of collision.

The second way that data may be interpreted wrongly is if the link is
temporarily interrupted and, by chance, a correct digest is received for these
incorrect data.  The digest appears to be sufficiently chaotic, so the
probability of this ought to be 1 in ~4*10^9, regardless of the nature of the
data or the interruption.  So the cable would have to be disconnected and
reconnected on the order of a billion times for this to be likely to occur even
once, provided that SOS signals are sent with reasonable frequency.

Finally, two identical starting digests may be generated for different sessions,
and then devices from different sessions may be connected and interpret received
messages as valid.  First of all, this concern is not particularly pressing, as
it is not common.  Secondly, its odds of occurring are 1 in ~4*10^9, assuming
that the random number generator is good.  Even if it were only somewhat good,
the odds would still be extremely low.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef LINK_MESSAGER_H_INCLUDED
#define LINK_MESSAGER_H_INCLUDED

//The number of bytes used to store data received from the link.  I'm not
//certain what the precise allowable lower bound on this is; it should be ~256.
//I chose not to sweat the last ~20 bytes.
#define LINK_BUFFER_SIZE 275
//The number of bytes used to store the session ID
#define SESSION_ID_SIZE 16
//The amount of time between repeated messages to start a session
#define SESSION_START_DELAY 500

//A buffer storing data received from the link.  LinkBuffer has its own
//abstraction methods in order to make code that retrieves data from the buffer
//easier to read and write.
//281 bytes
typedef struct {
	//Data received from the link.  It is a ring buffer.
	char data[LINK_BUFFER_SIZE];
	//The position in data of the marker, indicating the beginning of a message
	short markerPos;
	//The position in data of the next character to be read
	short readPos;
	//The position in data of the next character to be written
	short writePos;
} LinkBuffer;

//Data regarding a single session with a particular calculator
//323 bytes
typedef struct {
	LinkBuffer buffer;
	//An ID that is supposed to uniquely represent this session
	char sessionId[SESSION_ID_SIZE];
	//Whether startSession or joinSession has been called on this
	char connectionAttempted;
	//A random value used in establishing a session
	long randomValue;
	//The time when data were last sent in an attempt to start a new session
	long timeLastSentSession;
	//Whether any bytes have been received from the link
	char byteReceived;
	//The starting digest value used
	long startDigest;
	//The number of bytes that have matched the session ID.  Used when sending
	//an "SOS" signal to reset the read position.
	char sosBytes;
	//The current digest value for this message
	long curDigest;
	//The digest value indicated by receieved data
	char indicatedDigest[4];
	//The number of bytes that have been read of the digest value indicated by
	//receieved data
	char indicatedDigestBytes;
	//The size of the message currently being read, or a negative number if a
	//message is not currently being read
	char messageSize;
	//The position in the current message that is being read, if one is
	//currently being read
	char byteInMessage;
} LinkData;

//Stores a message that was received from or is to be sent through the link
typedef struct {
	//The contents of this message
	void *message;
	//The length of this message
	short size;
} Message;

//Frees the contents of the specified Message AND the pointer to the Message
//itself (it would be pretty trivial if it only did the former)
void freeMessage(Message *message) {
	free(message->message);
	free(message);
}

//Frees the contents of the specified Message AND the pointer to the Message
//itself, unless the message is NULL
void freeMessageIfNotNull(Message *message) {
	if (message != NULL)
		freeMessage(message);
}

//Returns a new message with a direct reference to the specified contents
Message *newMessage(void *message, short size) {
	Message *m = (Message *)malloc_throw(sizeof(Message));
	m->message = message;
	m->size = size;
	return m;
}

//Initializes the specified LinkBuffer
void initLinkBuffer(LinkBuffer *buffer) {
	memset(buffer, 0, sizeof(LinkBuffer));
}

//Returns a new LinkData for managing data received from the link
LinkData *newLinkData() {
	LinkData *ld = (LinkData *)malloc_throw(sizeof(LinkData));
	memset(ld, 0, sizeof(LinkData));
	initLinkBuffer(&(ld->buffer));
	ld->messageSize = -1;
	//Use a particular starting digest
	ld->startDigest = 1093212432l;
	//Use a false session ID used before the session ID is properly set
	memcpy(ld->sessionId,
		   (char[SESSION_ID_SIZE])
				{103, -128, -121, -71, 107, 10, -35, 30,
				 41, -111, 75, 49, -77, -29, -62, -19},
		   SESSION_ID_SIZE);
	return ld;
}

//Frees the contents of the specified LinkData (but not the pointer to the
//LinkData itself)
#define freeLinkData(x)

//Flushes all data that can be received from the link
static void flushLinkBuffer(LinkBuffer *lb) {
	char buffer[127];
	unsigned short numRead;
	short writePos = lb->writePos;
	char *data = lb->data;
	short bytesToRead = lb->readPos - writePos - 1;
	short firstSize;
	
	if (bytesToRead < 0)
		bytesToRead += LINK_BUFFER_SIZE;
	if (bytesToRead == 0)
		return;
	
	numRead = OSReadLinkBlock(buffer, min(bytesToRead, 127));
	//if (numRead > 0)
	//	printf("RCV %d\n", numRead);
	if (writePos + numRead <= LINK_BUFFER_SIZE)
		memcpy(data + writePos, buffer, numRead);
	else {
		firstSize = LINK_BUFFER_SIZE - writePos;
		memcpy(data + writePos, buffer, firstSize);
		memcpy(data, buffer + firstSize, numRead - firstSize);
	}
	
	lb->writePos = (writePos + numRead) % LINK_BUFFER_SIZE;
}

//Marks the current read position of the data
static void mark(LinkBuffer *lb) {
	lb->markerPos = lb->readPos;
}

//Returns to the last marker.  Assumes that the marker was not placed more than
//or exactly LINK_BUFFER_SIZE bytes ago.
static void returnToMarker(LinkBuffer *lb) {
	lb->readPos = lb->markerPos;
}

/*
//Reads a byte from the link buffer
static char readCharFromLinkBuffer(LinkBuffer *lb) {
	char val = lb->data[lb->readPos];
	if (++(lb->readPos) == LINK_BUFFER_SIZE)
		lb->readPos = 0;
	return val;
}
*/

//Skips the specified number of bytes in the specified LinkBuffer
static void skipBytesInLinkBuffer(LinkBuffer *lb, short numBytes) {
	lb->readPos = (lb->readPos + numBytes) % LINK_BUFFER_SIZE;
}

//Reads the specified number of bytes from the link buffer into the specified
//buffer without changing the current read position.  Assumes that that number
//of bytes is present.
static void peekAtBytesFromLinkBuffer(LinkBuffer *lb,
									  void *buffer,
									  short numBytes) {
	short readPos = lb->readPos;
	short firstSize;
	
	if (readPos + numBytes <= LINK_BUFFER_SIZE)
		memcpy(buffer, lb->data + readPos, numBytes);
	else {
		firstSize = LINK_BUFFER_SIZE - readPos;
		memcpy(buffer, lb->data + readPos, firstSize);
		memcpy(buffer + firstSize, lb->data, numBytes - firstSize);
	}
}

//Reads the specified number of bytes from the link buffer into the specified
//buffer.  Assumes that that number of bytes is present.
static void readBytesFromLinkBuffer(LinkBuffer *lb,
									void *buffer,
									short numBytes) {
	peekAtBytesFromLinkBuffer(lb, buffer, numBytes);
	skipBytesInLinkBuffer(lb, numBytes);
}

//Returns the number of bytes currently stored in the specified LinkBuffer after
//the current read position.
static short bytesInLinkBuffer(LinkBuffer *lb) {
	short result = lb->writePos - lb->readPos;
	return result >= 0 ? result : result + LINK_BUFFER_SIZE;
}

//Appends the specified byte to the specified digest value, and returns the
//result
static long digest(long val, char c) {
	//1485202837 and 597365980 were selected randomly (with the condition that
	//the former be odd)
	return (val + c) * 1485202837l + 597365980l;
}

//Digests the specified data
static long digestBytes(void *data, short size, long startingValue) {
	short i;
	long val = startingValue;
	for(i = 0; i < size; i++)
		val = digest(val, *((char *)data++));
	return val;
}

//Sends the specified data as a message.  Returns whether it was successful.
//Note that only 127 bytes may be sent in a message; longer messages must be
//broken up.
static char sendMessage(LinkData *ld, void *message, char size) {
	char toSend[140];
	long d;
	
	//if (OSLinkTxQueueInquire() < size + 5)
	//	return 0;
	
	//printf("SND %d\n", size + 5);
	
	toSend[0] = size;
	memcpy(toSend + 1, message, size);
	d = digestBytes(message, size, ld->startDigest);
	memcpy(toSend + size + 1, &d, 4);
	return !OSWriteLinkBlock(toSend, size + 5);
}

//Returns the next message available if there is one and NULL otherwise.  All
//transmissions are checked for message integrity.
Message *receiveMessage(LinkData *data) {
	short i;
	LinkBuffer *lb = &(data->buffer);
	char c;
	short bytes;
	Message *message;
	char *sessionId;
	char sosBytes;
	char nextSessionIdChar;
	char startSessionIdChar;
	char messageSize;
	long curDigest;
	char byteInMessage;
	char buffer[LINK_BUFFER_SIZE];
	char indicatedDigestBytes;
	char success;
	short m;
	short bytesToSkip;
	long startDigest;
	
	flushLinkBuffer(lb);
	bytes = bytesInLinkBuffer(lb);
	if (bytes <= 0)
		return NULL;
	
	sessionId = data->sessionId;
	sosBytes = data->sosBytes;
	nextSessionIdChar = sessionId[sosBytes];
	startSessionIdChar = *sessionId;
	messageSize = data->messageSize;
	byteInMessage = data->byteInMessage;
	curDigest = data->curDigest;
	indicatedDigestBytes = data->indicatedDigestBytes;
	startDigest = data->startDigest;
	m = -1;
	bytesToSkip = -1;
	
	peekAtBytesFromLinkBuffer(lb, buffer, bytes);
	
	data->byteReceived = 1;
	for(i = 0; i < bytes; i++) {
		c = buffer[i];
		//printf("%d\n", c);
		if (c == nextSessionIdChar) {
			if (++sosBytes == SESSION_ID_SIZE) {
				//The next byte should be the beginning of a message (or of
				//another SOS message)
				//printf("SOS\n");
				sosBytes = 0;
				messageSize = -1;
				continue;
			}
			else
				nextSessionIdChar = sessionId[sosBytes];
		}
		else {
			sosBytes = 0;
			nextSessionIdChar = startSessionIdChar;
		}
		
		if (messageSize < 0) {
			if (c >= 0) {
				messageSize = c;
				//printf("MS %d\n", c);
				byteInMessage = 0;
				indicatedDigestBytes = 0;
				curDigest = startDigest;
				m = i + 1;
			}
		}
		else if (byteInMessage == messageSize) {
			if (indicatedDigestBytes == 0 && i <= bytes - 4) {
				//This is supposed to be the more common case
				if ((success = !memcmp(buffer + i, &curDigest, 4)))
					bytesToSkip = i + 4;
				else
					messageSize = -1;
			}
			else {
				data->indicatedDigest[indicatedDigestBytes++] = c;
				if (indicatedDigestBytes == 4) {
					if (memcmp(data->indicatedDigest, &curDigest, 4)) {
						success = 0;
						messageSize = -1;
					}
					else {
						success = 1;
						bytesToSkip = i + 1;
					}
				}
				else
					success = 0;
			}
			
			if (success) {
				//printf("GET\n");
				message = (Message *)malloc_throw(sizeof(Message));
				if (m < 0) {
					returnToMarker(lb);
					readBytesFromLinkBuffer(lb,
											message->message =
												malloc_throw(messageSize),
											message->size = messageSize);
					skipBytesInLinkBuffer(lb, 4);
				}
				else {
					memcpy(message->message = malloc_throw(messageSize),
						   buffer + m,
						   message->size = messageSize);
					skipBytesInLinkBuffer(lb, bytesToSkip);
				}
				
				data->messageSize = -1;
				data->sosBytes = 0;
				return message;
			}
		}
		else {
			curDigest = digest(curDigest, c);
			byteInMessage++;
		}
	}
	
	data->sosBytes = sosBytes;
	data->messageSize = messageSize;
	data->curDigest = curDigest;
	data->byteInMessage = byteInMessage;
	data->indicatedDigestBytes = indicatedDigestBytes;
	
	if (m >= 0) {
		skipBytesInLinkBuffer(lb, m);
		mark(lb);
		skipBytesInLinkBuffer(lb, bytes - m);
	}
	else
		skipBytesInLinkBuffer(lb, bytes);
	
	return NULL;
}

//Sends a synchronization message that is supposed to make sure that if data
//transfer was interrupted, the receiver gets back on track.  (The "SOS" is as
//in "save our ship".)
void sendSync(LinkData *data) {
	OSWriteLinkBlock(data->sessionId, SESSION_ID_SIZE);
}

//Starts a new session with a particular calculator.  Returns whether it was
//successful.  A session must be started before any messages are sent or
//received.  It ought to be called repeatedly until a session is successfully
//joined.
char startSession(LinkData *ld) {
	Message *message;
	long l;
	long l2;
	long t = time();
	
	if (!(ld->connectionAttempted)) {
		ld->connectionAttempted = 1;
		//Certain values in the unsigned long range are impossible, but this
		//isn't of concern
		ld->randomValue = 32768l * ((long)rand()) + ((long)rand());
		ld->timeLastSentSession = t;
		OSLinkClose();
		OSLinkOpen();
	}
	
	//See if the other calculator has sent a key
	message = receiveMessage(ld);
	
	//Send a key
	if (t > ld->timeLastSentSession + SESSION_START_DELAY || message != NULL) {
		ld->timeLastSentSession = t;
		sendSync(ld);
		sendMessage(ld, &(ld->randomValue), 4);
	}
	
	if (message != NULL) {
		if (message->size == 4) {
			memcpy(&l, message->message, 4);
			
			//Embed the keys into the session ID and start digest, which
			//uniquely identify the session
			ld->startDigest = (l ^ ld->randomValue);
			l2 = min(l, ld->randomValue);
			memcpy(ld->sessionId, &l2, 4);
			l2 = max(l, ld->randomValue);
			memcpy(ld->sessionId + 4, &l2, 4);
			//Leave the last SESSION_ID_SIZE - 8 bytes of ld->sessionId alone
			
			//short i;
			//for(i = 0; i < SESSION_ID_SIZE; i++)
			//	printf("%d\n", ld->sessionId[i]);
			sendSync(ld);
			freeMessage(message);
			return 1;
		}
		else
			freeMessage(message);
	}
	
	return 0;
}

//Aborts an attempt to start a session
//data->connectionAttempted should be true
#define abortSessionAttempt(x) OSLinkClose()

//Closes the current session
//data->connectionAttempted should be true
#define closeSession(x) OSLinkClose()

//Resets the state of the link.  It is intended to be called when the link
//cable is physically disconnected.
#define resetLink(x) OSLinkClose(); OSLinkOpen()

















#endif
