/*
Created on August 12, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file provides a menu system whereby the user may select one of a fixed
number of states for a set of variables and/or select one from a set of options.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef MENU_H_INCLUDED
#define MENU_H_INCLUDED

#include "draw.h"
#include "extgraph.h"
#include "key.h"
#include "time.h"

//The maximum number of items that may be added to a menu
#define MAX_MENU_ITEMS 20
//The minimum amount of space between menu items
#define MIN_SPACE_BETWEEN_ITEMS 4
//The distance that menu items with values are drawn from the center of the
//screen.  Such items are drawn with the descriptive text right-aligned and
//DISTANCE_FROM_CENTER pixels left of the center and with the option text left-
//aligned and DISTANCE_FROM_CENTER pixels right of the center.
#define DISTANCE_FROM_CENTER 3

//Stores a single menu item.  Its fields should not be accessed outside this
//file.
typedef struct {
	//The message associated with this
	const char *message;
	//The acceptable values associated with this, if this stores items
	const char **values;
	//The number of values associated with this, or 0 if this does not store
	//items
	short numValues;
	//The index of the value that is currently selected, if this stores items
	short selected;
	//Whether this option can be "clicked", causing the user to leave the menu,
	//if numValues is 0
	char isClickable;
	//The font of this: F_4x6, F_6x8, or F_8x10
	short font;
} MenuItem;

//Represents data relevant to a menu.  Its fields should not be accessed outside
//this file.
typedef struct {
	//The items in this
	MenuItem items[MAX_MENU_ITEMS];
	//The space in pixels after each item in this.  The space for the last item
	//is ignored.  The array is parallel to items.
	short spaceAfter[MAX_MENU_ITEMS];
	//The number of items in this
	short numItems;
	//Whether the ESC button cancels the menu
	char escapeCancels;
} Menu;

//Returns a new MenuItem containing only the indicated message.  isClickable
//indicates whether it can be "clicked", causing the user to leave the menu
MenuItem *newMenuItem(const char *message, char isClickable, short font) {
	MenuItem *item = (MenuItem *)malloc_throw(sizeof(MenuItem));
	item->message = message;
	item->numValues = 0;
	item->font = font;
	item->isClickable = isClickable;
	return item;
}

//Returns a new MenuItem containing the indicated descriptive text (the
//"message" parameter) and menu options (the "values" parameter)
MenuItem *newMenuItemWithValues(const char *message,
								const char **values,
								short numValues,
								short selected,
								short font) {
	MenuItem *item = (MenuItem *)malloc_throw(sizeof(MenuItem));
	item->message = message;
	item->values = values;
	item->numValues = numValues;
	item->selected = selected;
	item->font = font;
	return item;
}

//Returns a new Menu
Menu *newMenu(char escapeCancels) {
	Menu *menu = (Menu *)malloc_throw(sizeof(Menu));
	menu->numItems = 0;
	menu->escapeCancels = escapeCancels;
	return menu;
}

//Frees the contents of the specified Menu (but not the pointer to the Menu
//itself)
#define freeMenu(x)

//Adds the specified MenuItem to the specified Menu, with the specified amount
//of space after it in pixels.  Normally, this should be 0.  If this is the last
//item in the menu, spaceAfter is ignored.  The Menu takes control of the
//MenuItem, becoming responsible for freeing it.
void addMenuItem(Menu *menu, MenuItem *item, short spaceAfter) {
	memcpy(menu->items + menu->numItems, item, sizeof(MenuItem));
	menu->spaceAfter[menu->numItems] = spaceAfter + MIN_SPACE_BETWEEN_ITEMS;
	(menu->numItems)++;
	free(item);
}

//Inverts the specified item, given its top y coordinate
static void drawMenuItem(void *buffer, MenuItem *item, short y) {
	const char *message = item->message;
	short font = item->font;
	short w = DrawStrWidth(message, item->font);
	const char *value;
	
	setDrawBuffer(buffer);
	
	FontSetSys(font);
	if (item->numValues > 0) {
		value = item->values[item->selected];
		DrawStr(80 - w - DISTANCE_FROM_CENTER, y, message, A_XOR);
		DrawStr(80 + DISTANCE_FROM_CENTER - 1, y, value, A_XOR);
	}
	else
		DrawStr((160 - w) / 2, y, message, A_XOR);
	
	PortRestore();
}

//Inverts a selection, given a tight rectangle representing the selection
void selectRect(void *buffer, short x1, short y1, short x2, short y2) {
	FastFillRect(buffer, x1 - 2, y1 - 2, x2 + 2, y2 + 2, A_XOR);
}

//Inverts the selection for the specified item, given its top y coordinate
void selectItem(void *buffer, MenuItem *item, short y) {
	short w;
	char font = item->font;
	
	if (item->numValues > 0)
		selectRect(buffer,
				   80 + DISTANCE_FROM_CENTER,
				   y,
				   80 + DISTANCE_FROM_CENTER +
						DrawStrWidth(item->values[item->selected], font) - 1,
				   y + fontHeight(font));
	else {
		w = DrawStrWidth(item->message, font);
		selectRect(buffer,
				   80 - w / 2, y,
				   80 + (w + 1) / 2, y + fontHeight(font));
	}
}

//Same as presentMenu, but repeatedly performs the specified action while
//waiting for the user to enter input, passing the specified data to the method,
//and exits, returning NULL, if action ever returns false.  It is guaranteed to
//be called at least once.  Automatically switches to menu mode.
short *presentMenuWithAction(Menu *menu,
							 long delay,
							 char (*action)(void *),
							 void *data) {
	short i, j, k;
	short y;
	
	short h = 0;
	short numItems = menu->numItems;
	short *spaceAfter = menu->spaceAfter;
	MenuItem *items = menu->items;
	MenuItem *item;
	MenuItem *curItem;
	short curItemIndex;
	short curItemY;
	short yo;
	unsigned short key;
	short *output;
	short numItemsWithValues;
	short bottom;
	char buffer[LCD_SIZE];
	
	switchToMenuMode();
	
	clearDrawBuffer(buffer);
	
	numItemsWithValues = 0;
	bottom = -1;
	for(i = 0, item = items; i < numItems; i++, item++) {
		bottom = h;
		h += fontHeight(item->font);
		if (i < numItems - 1)
			h += spaceAfter[i];
		if (item->numValues > 0)
			numItemsWithValues++;
	}
	yo = (100 - h) / 2;
	bottom += yo;
	
	//Draw the menu, and compute the initial values for curItem, curItemIndex,
	//and curItemY
	ClrScr();
	y = yo;
	curItem = NULL;
	curItemIndex = -1;
	curItemY = -1;
	for(i = 0, item = items; i < numItems; i++, item++) {
		drawMenuItem(buffer, item, y);
		if (curItemIndex < 0 && (item->numValues > 0 || item->isClickable)) {
			curItem = item;
			curItemIndex = i;
			curItemY = y;
		}
		y += fontHeight(item->font) + spaceAfter[i];
	}
	
	sleep(delay);
	flushKeyQueue();
	while(1) {
		//Select the current position
		selectItem(buffer, curItem, curItemY);
		
		//Display the buffer
		draw89Size(buffer);
		
		//Respond to input
		idle();
		while(1) {
			do {
				if (!action(data))
					return NULL;
			}
			while (!(key = nextKey()));
			
			if (key == KEY_ESC && menu->escapeCancels)
				return NULL;
			else if (key == KEY_ENTER) {
				if (curItem->numValues > 0) {
					key = KEY_RIGHT;
					break;
				}
				else
					goto BreakOuter;
			}
			else if (key == KEY_UP || key == KEY_DOWN ||
					 ((key == KEY_LEFT || key == KEY_RIGHT) &&
					  curItem->numValues > 0))
				break;
		}
		
		//Deselect the current position
		selectItem(buffer, curItem, curItemY);
		
		if (key == KEY_LEFT || key == KEY_RIGHT) {
			drawMenuItem(buffer, curItem, curItemY);
			if (key == KEY_LEFT) {
				if ((curItem->selected)-- <= 0)
					curItem->selected = curItem->numValues - 1;
			}
			else if (++(curItem->selected) >= curItem->numValues)
				curItem->selected = 0;
			drawMenuItem(buffer, curItem, curItemY);
		}
		else {
			curItemIndex = curItemIndex;
			do {
				if (key == KEY_UP) {
					if (curItemIndex-- <= 0) {
						curItemIndex = numItems - 1;
						curItemY = bottom;
					}
					else
						curItemY -= fontHeight(items[curItemIndex].font) +
							spaceAfter[curItemIndex];
				}
				else {
					curItemY += fontHeight(curItem->font) +
						spaceAfter[curItemIndex];
					if (++curItemIndex >= numItems) {
						curItemIndex = 0;
						curItemY = yo;
					}
				}
				curItem = items + curItemIndex;
			}
			while (curItem->numValues == 0 && !(curItem->isClickable));
		}
	}
BreakOuter:	
	output = (short *)malloc_throw((numItemsWithValues + 1) * sizeof(short));
	j = 1;
	k = 0;
	for(i = 0, item = items; i < numItems; i++, item++) {
		if (i == curItemIndex)
			output[0] = k;
		
		if (item->numValues > 0)
			output[j++] = item->selected;
		else if (item->isClickable)
			k++;
	}
	return output;
}

//An action for presentMenuWithAction that does nothing
char emptyAction(void *data) {
	return 1;
}

//Presents the specified menu to the user, waiting until the user has performed
//all desired changes.  delay indicates the amount of time to delay before
//accepting user input.  Returns an array containing the item number that was
//"clicked" to exit the menu, followed by the indices of the values given by the
//user in the menu items in the menu with values.  The latter is in the order in
//which the items were added, and the former is the item number relative to
//those that are "clickable".  Instead returns NULL if the menu was canceled by
//pressing escape.  Automatically switches to menu mode.
short *presentMenu(Menu *menu, long delay) {
	return presentMenuWithAction(menu, delay, emptyAction, NULL);
}

















#endif
