/*
Originally created on August 15, 2006 by William Jacobs.  Substantially
rewritten beginning on June 30, 2007 by William Jacobs.

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file provides a synchronized higher-level messaging system that utilizes
linkmessager.h.  It assumes that each calculator sends one message for each one
that it receives.  That is to say, the client may only receive at most one more
message than is sent and vice versa.  It automatically displays a linking error
message when appropriate and automatically resumes when the user reconnects the
link.  It is intended to simplify linking and to provide the reconnect
functionality.  Note that the reconnection functionality is fairly expensive
with regard to program size and complexity.  messagehandler.h inherits from
linkmessager.h the assurance of data integrity and the expense with regard to
transfer rate.  See linkmessager.h for more details.

Each message that is sent must be at least one character long, and the first
byte, interpreted as a character, must be at least 0 and at most 49, so that the
message handler can store additional information by altering it.

Each message is sent with one of four possible states attached to it.  Messages
store their numbers mod 4, so that the first sent message is 0, the second is 1,
the third is 2, the fourth is 3, the fifth is 0, and so on.  When a calculator
has waited too long for a message, it repeatedly send an "SOS" signal followed
by copies of the last two messages sent.  After a while, it decides that the
link is broken and displays a message indicating that it has been disconnected.
When the link is broken, copies of the last two messages are repeatedly sent by
both calculators until one of the calculators has received the message for which
it had been waiting.  This calculator sends the message MHM_LINK_FIXED to
indicate that the link has been fixed.  (Note that the link is considered to be
fixed if either calculator has received the message for which it had been
waiting; otherwise, such a calculator may never be able to compute and send the
message for which the other was waiting.)  Upon receiving the message
MHM_LINK_FIXED, the other calculator responds with the same message.  Then, both
calculators display a message indicating that the connection was restored and
wait until a player presses ENTER or ESC, upon which either the message
MHM_RESUME or MHM_QUIT is sent, and messaging returns to normal.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef MESSAGE_HANDLER_H_INCLUDED
#define MESSAGE_HANDLER_H_INCLUDED

#include "draw.h"
#include "key.h"
#include "linkmessager.h"
#include "time.h"

//The amount of time to wait before resending old messages
#define RESEND_DELAY 1000
//The amount of time until the first resending messages when waiting to receive
//a message
#define FIRST_RESEND 500
//The amount of time to wait until displaying a message indicating that the link
//is broken and trying to re-establish the connection
#define LINK_BROKEN_DELAY 3000

//Data used for simplified handling of messages.  It assumes that each
//calculator sends one message for each message that the other sends.  It
//automatically handles broken connections, allowing them to be repaired simply
//by reconnecting the calculators or aborted.  Its fields should not be accessed
//outside this file.
//353 bytes, plus the size of stored messages
typedef struct {
	LinkData *linkData;
	//The last two sent messages, so that sent[i] indicates a message numbered
	//i, mod 2.  Some elements may be NULL.
	Message *sent[2];
	//The last four received messages, so that received[i] indicates a message
	//numbered i, mod 4.  At most two elements are not NULL: the next message
	//to be retrieved and the one after that.
	Message *received[4];
	//The number of messages that have been sent mod 4
	char sentMod4;
	//The number of messages that have been received mod 4
	char receivedMod4;
} MessageHandler;

//Special messages sent by the message handling system.  MHM_LINK_FIXED is sent
//when a calculator's link has been repaired.  MHM_RESUME and MHM_QUIT are used
//to indicate that a user pressed ENTER or ESC after a link was broken, then
//repaired.  When converted from a character to an unsigned character, each must
//be at least 50 * 4 = 200.
enum MessageHandlerMessageTypes {MHM_LINK_FIXED = -3, MHM_RESUME = -2, MHM_QUIT = -1};

//Frees the contents of the specified MessageHandler (but not the pointer to the
//MessageHandler itself) (public function)
void freeMessageHandler(MessageHandler *mh) {
	short i;
	
	freeLinkData(mh->linkData);
	free(mh->linkData);
	
	for(i = 0; i < 2; i++)
		freeMessageIfNotNull(mh->sent[i]);
	for(i = 0; i < 4; i++)
		freeMessageIfNotNull(mh->received[i]);
}

//Returns a new MessageHandler (public function)
MessageHandler *newMessageHandler() {
	MessageHandler *mh = (MessageHandler *)malloc_throw(sizeof(MessageHandler));
	memset(mh, 0, sizeof(MessageHandler));
	mh->linkData = newLinkData();
	return mh;
}

//Starts a new session with a particular calculator.  Returns whether it was
//successful.  A session must be started before any messages are sent or
//received.  It ought to be called repeatedly until a session is successfully
//joined. (public function)
#define handlerStartSession(x) startSession((x)->linkData)

//Aborts an attempt to start a session (public function)
#define handlerAbortSessionAttempt(x) abortSessionAttempt((x)->linkData)

//Closes the current session (public function)
#define handlerCloseSession(x) closeSession((x)->linkData)

//Draws a message indicating that the connection has been reestablished, and
//sleeps for a short time
void drawConnectionReestablished() {
	drawMessageAndSleep("Connection\nre-established`Press ENTER to resume`"
					    "Press ESC to exit");
}

//Draws a message indicating that the link is broken, and sleeps for a short
//time
void drawLinkBroken() {
	drawMessageAndSleep("Link broken`Check your connection`"
					    "The game should automatically resume\n"
					    "when you reconnect your cable`"
					    "Press ESC or ON to abort the game");
}

//Returns the number of a message a message mod 4, based on its first byte
char messageNumMod4(char c) {
	return (((unsigned char)c) / 50) % 4;
}

//Alters the first byte of the message to be the number 0 to 49 it encodes
inline void handlerClearInfo(Message *message) {
	((unsigned char *)(message->message))[0] %= 50;
}

//Alters the first byte of the message to encode the number of the message mod 4
void handlerAddInfo(Message *message, char mod4) {
	((unsigned char *)(message->message))[0] =
		((((unsigned char *)(message->message))[0]) % 50) +
			50 * (unsigned char)mod4;
}

//Sends the specified one-byte message
void handlerSendSimpleMessage(MessageHandler *mh, char type) {
	sendMessage(mh->linkData, &type, 1);
}

//The return type of handlerWaitForMessage.  Stores the message retrieved and
//the amount of time spent waiting to receive a message.
typedef struct {
	//The message that was retrieved, or NULL if no message was retrieved
	Message *message;
	//The total amount of time spent waiting to receive a message since the last
	//message was received
	long time;
} MessageAndTime;

//Flags used in handlerReceiveMessage0 indicating the state of a calculator with
//regard to its link.  MHLS_BOTH_BROKEN indicates that a required message has
//been missing for so long that a "link broken" message is displayed.
//MHLS_OPPONENT_BROKEN indicates that such a message was subsequently received.
//MHLS_BOTH_FIXED indicates that a message of the type MHM_LINK_FIXED was
//received.  It returns to MHLS_NORMAL when someone subsequently presses ENTER.
enum MessageHandlerLinkState
	{MHLS_NORMAL = 0,
	 MHLS_BOTH_BROKEN = 1,
	 MHLS_OPPONENT_BROKEN = 2,
	 MHLS_BOTH_FIXED = 3};

//Function for receiving messages, with a couple of options.  wait is whether to
//wait until a message is available.  returnMessage is whether to remove and
//return a retrieved message from the state of the MessageHandler, rather than
//just storing retrieved messages.  Returns NULL if the user quit and a
//MessageAndTime object with a NULL message field if no message was retrieved.
MessageAndTime *handlerReceiveMessage0(MessageHandler *mh,
									   long timeElapsed,
									   char wait,
									   char returnMessage) {
	short i;
	MessageAndTime *mat =
		(MessageAndTime *)malloc_throw(sizeof(MessageAndTime));
	
	//Both or neither of these is MHF_NORMAL
	char state = MHLS_NORMAL;
	
	char displayingLinkBroken = 0;
	char displayingConnectionReestablished = 0;
	Message *message;
	char c;
	char mod4;
	long start = time() - timeElapsed;
	long nextResend = start + FIRST_RESEND;
	long t;
	unsigned short key;
	
	mat->message = NULL;
	mat->time = 0;
	while (1) {
		if (timerExpired()) {
			free(mat);
			return NULL;
		}
		t = time();
		
		//Receive
		while ((message = receiveMessage(mh->linkData)) != NULL) {
			c = ((char *)(message->message))[0];
			switch(c) {
				case MHM_LINK_FIXED:
					if (state <= MHLS_BOTH_BROKEN)
						handlerSendSimpleMessage(mh, MHM_LINK_FIXED);
					state = MHLS_BOTH_FIXED;
					freeMessage(message);
					break;
				case MHM_RESUME:
					state = MHLS_NORMAL;
					start = t;
					nextResend = t + RESEND_DELAY;
					freeMessage(message);
					break;
				case MHM_QUIT:
					drawOpponentQuit();
					free(mat);
					freeMessage(message);
					idle();
					while (nextKey() != KEY_ESC);
					return NULL;
				default:
					mod4 = messageNumMod4(c);
					
					if (mod4 == mh->receivedMod4 ||
							mod4 == (mh->receivedMod4 + 1) % 4) {
						freeMessageIfNotNull(mh->received[mod4]);
						mh->received[mod4] = message;
						if (state == MHLS_BOTH_BROKEN &&
								mod4 == mh->receivedMod4) {
							state = MHLS_OPPONENT_BROKEN;
							nextResend = t;
						}
					}
					else
						freeMessage(message);
			}
		}
		
		if (state == MHLS_NORMAL) {
			mod4 = mh->receivedMod4;
			message = mh->received[mod4];
			if (message != NULL) {
				if (returnMessage) {
					mh->received[mod4] = NULL;
					mh->receivedMod4 = (mod4 + 1) % 4;
					handlerClearInfo(message);
					mat->message = message;
				}
				return mat;
			}
		}
		
		if (!wait)
			return mat;
		
		mat->time = t - start;
		
		//Resend
		if (t >= nextResend && state != MHLS_BOTH_FIXED) {
			if (state == MHLS_NORMAL &&
					t > start + LINK_BROKEN_DELAY)
				state = MHLS_BOTH_BROKEN;
			
			resetLink(mh->linkData);
			sendSync(mh->linkData);
			if (state != MHLS_OPPONENT_BROKEN) {
				for(i = 0; i < 2; i++) {
					message = mh->sent[i];
					if (message != NULL) {
						handlerAddInfo(message, (i != (mh->sentMod4 % 2)) ?
													(mh->sentMod4 + 3) % 4 :
													(mh->sentMod4 + 2) % 4);
						sendMessage(mh->linkData,
									message->message,
									message->size);
					}
				}
			}
			else
				handlerSendSimpleMessage(mh, MHM_LINK_FIXED);
			
			nextResend = t + RESEND_DELAY;
		}
		
		if (state != MHLS_NORMAL) {
			//Draw (must appear before checking for key presses, so that
			//switchToMenuMode is called)
			if (state != MHLS_BOTH_FIXED) {
				if (!displayingLinkBroken) {
					drawLinkBroken();
					displayingLinkBroken = 1;
					displayingConnectionReestablished = 0;
				}
			}
			else if (!displayingConnectionReestablished) {
				drawConnectionReestablished();
				displayingConnectionReestablished = 1;
				displayingLinkBroken = 0;
			}
			
			//Check key presses
			while ((key = nextKey())) {
				if (key == KEY_ESC) {
					handlerSendSimpleMessage(mh, MHM_QUIT);
					free(mat);
					return NULL;
				}
				else if (key == KEY_ENTER && state == MHLS_BOTH_FIXED) {
					handlerSendSimpleMessage(mh, MHM_RESUME);
					if (!wait)
						return mat;
					state = MHLS_NORMAL;
					start = t;
					nextResend = t + RESEND_DELAY;
				}
			}
		}
	}
}

//The return type for handlerReceiveMessage
typedef struct {
	//The message that was received, or NULL if no message was available
	Message *message;
	//Whether the user chose to abort the game as a result of a failed
	//connection
	char quit;
} ReceiveMessageResult;

//Returns a ReceiveMessageResult corresponding to the next message available.
//If no message is available, the return value's message field will be NULL.
//All transmissions are checked for message integrity, using the functionality
//provided by linkmessage.h.  It assumes that each calculator sends one message
//for each message that the other sends.  Never returns NULL. (public function)
ReceiveMessageResult *handlerReceiveMessage(MessageHandler *mh) {
	ReceiveMessageResult *rmr =
		(ReceiveMessageResult *)malloc_throw(sizeof(ReceiveMessageResult));
	MessageAndTime *mat = handlerReceiveMessage0(mh, 0, 0, 1);
	
	if (mat != NULL) {
		rmr->message = mat->message;
		rmr->quit = 0;
		free(mat);
	}
	else {
		rmr->message = NULL;
		rmr->quit = 1;
	}
	return rmr;
}

//Waits until a message is available, and then returns it and the amount of time
//it took to retrieve it.  timeElapsed indicates the amount of time the program
//has spent looking for a message of a particular description (as determined by
//the client).  This way, the message handler can know how soon to display a
//"link broken" message.  All transmissions are checked for message integrity.
//It assumes that each calculator sends one message for each message that the
//other sends and properly generates error messages when a message that ought to
//have been recieved has not been.  Returns a MessageAndTime object with a NULL
//message field if it failed (i.e. the link was broken and the user opted to
//quit). (public function)
MessageAndTime *handlerWaitForMessage(MessageHandler *mh, long timeElapsed) {
	MessageAndTime *mat = handlerReceiveMessage0(mh, timeElapsed, 1, 1);
	if (mat == NULL) {
		mat = (MessageAndTime *)malloc_throw(sizeof(MessageAndTime));
		mat->message = NULL;
		mat->time = 0;
	}
	return mat;
}

//Sends the specified data as a message.  It assumes that each calculator sends
//one message for each message that the other sends and properly generates error
//messages when a message that ought to have been recieved has not been.  Under
//these circumstances, it does its best to allow the users to resume the game by
//reconnecting the calculators.  It also assumes that size > 0 &&
//((char *)message)[0] >= 0 && ((char *)message)[0] < 50.  Returns whether it
//was successful.  It takes control of message, using it destructively and
//becoming responsible for freeing it.  If this returns false, this means the
//link is broken and cannot be reestablished.  Note that only 125 bytes may be
//sent in a message; longer messages must be broken up. (public function)
char handlerSendMessage(MessageHandler *mh, void *message, char size) {
	char mod4 = mh->sentMod4;
	MessageAndTime *mat;
	Message *message2;
	
	mat = handlerReceiveMessage0(mh, 0, 0, 0);
	if (mat == NULL)
		return 0;
	free(mat);
	message2 = newMessage(message, size);
	handlerAddInfo(message2, mod4);
	sendMessage(mh->linkData, message, size);
	
	freeMessageIfNotNull(mh->sent[mod4 % 2]);
	mh->sent[mod4 % 2] = message2;
	mh->sentMod4 = (mod4 + 1) % 4;
	return 1;
}










#endif
