/*
Created on August 9, 2006 by William Jacobs

This file was originally created as a part of the CalcChu project, whose aim was
to produce a 68K calculator game based on ChuChu Rocket! that allows for two-
player games using a link cable.

This file contains code that permits simplified execution of a timed game,
without specifying the user interface.

The comments in this file often use the term "game time".  This indicates the
time spent advancing the state of the game and excludes time spent displaying a
message.

To keep the games displayed on linked calculators identical, a random number
generator seeded with the same value is used on each calculator, the usage of
which is carefully controlled.  Thus, seemingly innocent changes to some
algorithms may result in incompatibility with alternative versions of the game.
The rules for executing the game are as follows, with uses of the random number
generator indicated:

- Critters hatched using certain probabilities are hatched as indicated by the
comments in game.h.

- When selecting a random event to be executed, an index for the random event is
selected by calling rngRandom(rng, NUM_EVENTS) until an index different than the
index of the previous event for the game (if any) is found.

- Special events that use the random number generator have their own rules for
using it.



This is provided under two licenses: one non-commercial and non-restrictive, and
one that is non-restrictive and allows for commercial usage, but excludes the
ExtGraph library.  The software is provided under the following two licenses:

License 1:

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software for any non-commercial purpose whatever, including the rights to
non-commercially use, copy, modify, merge, publish, distribute, and/or
sublicense the Software, and to permit persons to whom the Software is furnished
to do so, subject to the following conditions:

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



License 2:

The "ExtGraph library" refers to the ExtGraph software and associated
documentation files.  The "Software" is hereby taken to mean the code contained
in this and associated files for this software, excluding those which the
ExtGraph library contains, and associated documentation files.

Permission is hereby granted, free of charge, to any person obtaining a copy of
the Software to deal in the Software without restriction, including without
limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, to derive executable and/or
linkable software from the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

All references whatever to the ExtGraph library shall be omitted or replaced
with suitable substitutes.

The above notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef TIMED_GAME_H_INCLUDED
#define TIMED_GAME_H_INCLUDED

#include "draw.h"
#include "game.h"
#include "rng.h"
#include "time.h"

//10000 times the default amount of time between game ticks
#define DEFAULT_DELAY10000 750000l
//The number of arrows each player can place
#define NUM_ARROWS 3
//The amount of time that an arrow lasts
#define ARROW_TIME 10000l
//The amount of time that an event message is displayed
#define EVENT_MESSAGE_TIME 1200l
//The amount of time for which the message at the start of the game is displayed
#define START_MESSAGE_TIME 1500l
//The amount of time for which the message at the end of the game is displayed
#define END_MESSAGE_TIME 3000l
//The maximum amount of time by which to advance per update
#define MAX_UPDATE_INTERVAL 500l

//The amount of time left for an arrow at which the arrow begins to flash.  For
//best appearance, should be an even multiple of ARROW_FLASH TIME.
#define ARROW_FLASH_START 1000l
//The interval at which an arrow flashes
#define ARROW_FLASH_TIME 100l

//An enumeration of types of game messages displayed on the screen
enum MessageTypes {NORMAL_MESSAGE, START_MESSAGE, END_MESSAGE};
//The critters whose motion is paused.  Used as a field for special events.
enum PauseTypes {PAUSE_NONE, PAUSE_CATS, PAUSE_CATS_AND_MICE};

//Stores data pertaining to an in-game message
//10 bytes
typedef struct {
	//Whether this GameMessage object represents a message that is actually
	//being shown, as opposed to merely occupying space.  Note that no
	//guarantees are made regarding the other fields if this is false.
	char isShown;
	//The amount of time in which which the message will disappear.  Guaranteed
	//to be nonnegative.
	long time;
	//The string message
	const char *message;
	//The type of this message: NORMAL_MESSAGE, START_MESSAGE, or END_MESSAGE
	char type;
} GameMessage;

struct TIMED_GAME;

//Stores data pertaining to a special event
//22 bytes
typedef struct {
	//Whether this Event object represents a special event is actually being
	//executed, as opposed to merely occupying space.  Note that no guarantees
	//are made regarding the other fields if this is false.
	char exists;
	//Data specific to this special event
	void *data;
	//Function for starting this special event.  Assumes that this exists.
	void (*start)(struct TIMED_GAME *);
	//Function for stopping this special event and freeing any data indicated by
	//data.  Assumes that this exists.
	void (*stop)(struct TIMED_GAME *);
	//Function for drawing extra stuff in the background.  (The buffer is
	//guaranteed to have BORDER rows above and below the usual board.)
	void (*draw)(void *, struct TIMED_GAME *);
	//For what types of critters motion is paused for this.  One of PAUSE_NONE,
	//PAUSE_CATS, and PAUSE_CATS_AND_MICE.  For critters that are paused, 0 is
	//used as the fraction of a tick that they have advanced to the next space.
	char pause;
	//The amount of game time that was remaining for the event when the last
	//game tick was executed
	long time;
} Event;

//Stores data used to maintain the timed game.  Its fields should not be
//accessed outside this file.
//3454 bytes
typedef struct TIMED_GAME {
	//The game for this
	Game *game;
	//Draw buffer for the board only; it excludes things such as critters.
	//Drawn to at load time.
	unsigned char boardDrawBuffer[BOARD_DRAW_SIZE];
	//The amount of game time until the next game tick (note that this excludes
	//time during which a message is displayed).  Guaranteed to be nonnegative.
	long timeUntilNextTick;
	//The last time this was updated
	long lastTime;
	//The amount of time between game ticks
	long delay;
	//Whether the game has been started, and time has been initialized
	char started;
	//The amount of time remaining in the game.  Guaranteed to be nonnegative.
	long timeRemaining;
	//The amount of time until player 1's arrows disappear.  The array is
	//parallel to game->p1Arrows.
	long p1ArrowTimes[MAX_ARROWS];
	//The amount of time until player 2's arrows disappear.  The array is
	//parallel to game->p2Arrows.
	long p2ArrowTimes[MAX_ARROWS];
	//The message that is currently being shown, if any
	GameMessage message;
	//Whether the user is player 1
	char isPlayer1;
	//Whether the game is completely over, including displaying the end message
	char gameOver;
	//Whether to draw the opponent's arrows.  This is set to false for the
	//"Invisible Arrows" special event
	char drawOpponentsArrows;
	//The speed modifier for the game times 100.  Does not include the speed
	//modifier for such things as the "Speed Up" special event.  Note that such
	//"additional" modifiers are not stored in TimedGame, but are factored into
	//delay.
	short speed100;
	
	//The method for updating the game by one game tick.  Usually, this is
	//advanceTickNormal, but if there is an event, this may be something special
	//associated with the event.  Returns whether a special event should
	//executed, and, if so, the player for whom it should be executed, 1 or 2.
	char (*advanceTick)(struct TIMED_GAME *);
	//Data for the current special event.
	Event event;
	//Whether this is the first tick that a particular special event is being
	//executed, and, if so, the player number for whom it was generated (1 or
	//2).  The value is maintained in updateTimedGame.
	char firstEventTick;
	//The index for the last event that was executed or the event that is
	//currently being executed.  It is negative if no event has been executed.
	short lastEventIndex;
	//The random number generator for this
	RNG *rng;
} TimedGame;

static char advanceTickNormal(TimedGame *tg) {
	return advanceTick(tg->game);
}

//Sets the speed of the specified TimedGame to be speed100 / 100 times the
//normal speed (as determined by tg->speed100)
static void setSpeed100(TimedGame *tg, short speed100) {
	tg->delay =
		max(DEFAULT_DELAY10000 / ((long)(tg->speed100) * (long)speed100), 1);
}

//Returns a new TimedGameData for the specified game
TimedGame *newTimedGame(Game *game,
						long duration,
						short speed100,
						char isPlayer1,
						RNG *rng) {
	TimedGame *tg = (TimedGame *)malloc_throw(sizeof(TimedGame));
	memset(tg, 0, sizeof(TimedGame));
	tg->game = game;
	tg->timeRemaining = duration;
	tg->isPlayer1 = isPlayer1;
	tg->advanceTick = advanceTickNormal;
	tg->drawOpponentsArrows = 1;
	tg->speed100 = speed100;
	tg->rng = rng;
	tg->lastEventIndex = -1;
	drawBoard(tg->boardDrawBuffer, game->board);
	setSpeed100(tg, 100);
	return tg;
}

//Frees the contents of the specified TimedGame (but not the pointer to the
//TimedGame itself)
#define freeTimedGame(x)

//Removes the specified arrow from the specified game
void destroyArrow(Arrow *arrow, Game *game) {
	game->boardData[arrow->pos].arrowDir = NO_DIRECTION;
	arrow->exists = 0;
}

//Helper function for decreaseArrowTimes
void decreaseArrowTimes0(Game *game,
						 Arrow *arrows,
						 long *arrowTimes,
						 long time) {
	short i;
	Arrow *arrow;
	
	for(i = 0, arrow = arrows; i < MAX_ARROWS; i++, arrow++) {
		if (arrow->exists && (arrowTimes[i] -= time) <= 0)
			destroyArrow(arrow, game);
	}
}

//Decrease the times associated with the arrows by the specified number of
//milliseconds.  Removes arrows that have expired.
void decreaseArrowTimes(TimedGame *tg, long time) {
	Game *game = tg->game;
	decreaseArrowTimes0(game, game->p1Arrows, tg->p1ArrowTimes, time);
	decreaseArrowTimes0(game, game->p2Arrows, tg->p2ArrowTimes, time);
}

//Drops an arrow at the specified position for the specified player.  Assumes
//that there is not a pit or rocket there.  Removes the oldest arrow if
//NUM_ARROWS of his arrows are already in place.  Returns whether anything
//happened (i.e. whether the arrow was successfully placed).
char dropArrow(TimedGame *tg, char isPlayer1, char pos, char dir) {
	short i;
	Game *game = tg->game;
	Arrow *arrows = isPlayer1 ? game->p1Arrows : game->p2Arrows;
	long *arrowTimes = isPlayer1 ? tg->p1ArrowTimes : tg->p2ArrowTimes;
	short earliest = -1;
	Arrow *arrow;
	long *arrowTime;
	long earliestTime = LONG_MAX;
	SpaceData *space = game->boardData + pos;
	char type = space->type;
	
	if (type != SPACE_NOTHING || space->arrowDir != NO_DIRECTION)
		return 0;
	
	//Find the earliest placed arrow and count the number of arrows, to see
	//which arrow, if any, should be removed
	for(i = 0, arrowTime = arrowTimes, arrow = arrows;
			i < MAX_ARROWS;
			i++, arrowTime++, arrow++) {
		if (arrow->exists) {
			if (*arrowTime < earliestTime) {
				earliest = i;
				earliestTime = *arrowTime;
			}
		}
		else
			break;
	}
	
	//Find the appropriate arrow
	if (i == NUM_ARROWS) {
		arrow = arrows + earliest;
		arrowTime = arrowTimes + earliest;
		destroyArrow(arrow, game);
	}
	
	//Alter the appropriate arrow
	game->boardData[pos].arrowDir = dir;
	arrow->exists = 1;
	arrow->pos = pos;
	arrow->dir = dir;
	arrow->isSmall = 0;
	*arrowTime = ARROW_TIME - (tg->delay - tg->timeUntilNextTick);
	
	return 1;
}

//Helper function for drawArrowsForTimedGame.  Draws the specified arrow if it
//exists and is not hidden due to arrow flashing.
void maybeDrawArrowForTimedGame(void *buffer,
								char isPlayer1,
								Arrow *arrow,
								long arrowTime) {
	if (arrow->exists &&
			(arrowTime > ARROW_FLASH_START ||
			 (((ARROW_FLASH_START - arrowTime) / ARROW_FLASH_TIME) % 2)))
		drawArrow(buffer, isPlayer1, arrow->pos, arrow->dir, arrow->isSmall);
}

//Similar to drawArrows, but uses arrow flashing
void drawArrowsForTimedGame(void *buffer, TimedGame *tg) {
	short i;
	
	for(i = 0; i < MAX_ARROWS; i++) {
		if (tg->drawOpponentsArrows || tg->isPlayer1)
			maybeDrawArrowForTimedGame(buffer,
									   1,
									   tg->game->p1Arrows + i,
									   tg->p1ArrowTimes[i]);
		if (tg->drawOpponentsArrows || !(tg->isPlayer1))
			maybeDrawArrowForTimedGame(buffer,
									   0,
									   tg->game->p2Arrows + i,
									   tg->p2ArrowTimes[i]);
	}
}

//Draws the background of the current state of the game, which at present is
//everything but the in-game message, to the specified buffer.  Assumes that the
//buffer has EXPANDED_DRAW_OFFSET garbagy elements at the beginning (see
//drawCritter).  Effectively clears the buffer first.
void drawTimedGameBackground(void *buffer, TimedGame *tg) {
	Game *game = tg->game;
	short frac;
	Event *e = &(tg->event);
	
	memset(buffer + BOARD_DRAW_SIZE, 0, LCD_SIZE - BOARD_DRAW_SIZE);
	memcpy(buffer, tg->boardDrawBuffer, BOARD_DRAW_SIZE);
	if (e->exists && !(tg->firstEventTick))
		e->draw(buffer, tg);
	drawRockets(buffer, game);
	drawArrowsForTimedGame(buffer, tg);
	frac = tg->message.isShown ? 0 :
		(short)max(0, min((MAX_FRAC + 0.5) -
						  ((MAX_FRAC * (float)tg->timeUntilNextTick) /
						   (float)tg->delay), MAX_FRAC - 1));
	drawCritters(buffer,
				 game,
				 e->pause == PAUSE_NONE ? frac : 0,
				 e->pause != PAUSE_CATS_AND_MICE ? frac : 0);
	drawScores(buffer, game);
	drawTime(buffer, (short)(tg->timeRemaining / 1000));
}

//Draws the foreground of the current state of the game, which for now is just
//the in-game message
void drawTimedGameForeground(void *buffer, TimedGame *tg) {
	if (tg->message.isShown)
		drawGameMessage(buffer, tg->message.message);
}

//Returns whether the game is completed
#define gameOver(x) (x)->gameOver

//Returns whether normal game input should currently be accepted.  Notably, this
//excludes input that might bring the user to a menu.
char allowInput(TimedGame *tg) {
	return !(tg->message.isShown);
}

//Sets the TimedGame to display the specified message for MESSAGE_TIME time
void postMessage(TimedGame *tg,
				 const char *str,
				 long duration,
				 char messageType) {
	GameMessage *message = &(tg->message);
	message->isShown = 1;
	message->message = str;
	message->time = duration;
	message->type = messageType;
}










//===== Code for game events =====
//Does nothing
void doNothingOnTimedGame(TimedGame *tg) {
	
}

//Does nothing
void drawNothing(void *buffer, TimedGame *tg) {
	
}

//Does nothing
char doNothingAdvanceEvent() {
	return 0;
}

//Destroys all critters in the specified timed game
void destroyCrittersForTimedGame(TimedGame *tg) {
	destroyCritters(tg->game);
}

//The fixed number of cats in cat mania
#define NUM_CATS_IN_CAT_MANIA 2

char advanceEventCatMania(TimedGame *tg) {
	short i, j;
	
	Game *game = tg->game;
	short hatchIndices[MAX_HATCHES];
	Hatch *hatches = game->board->hatches;
	short numHatches = game->board->numHatches;
	Hatch *hatch;
	short cats;
	Critter *critter;
	
	cats = numCats(game);
	
	//Try to get to NUM_CATS_IN_CAT_MANIA cats
	if (cats < NUM_CATS_IN_CAT_MANIA) {
		j = 0;
		for(i = 0; i < numHatches; i++) {
			if (!tg->game->boardData[hatches[i].pos].numCats)
				hatchIndices[j++] = i;
		}
		
		//Randomly select hatches from which to hatch
		while (cats < NUM_CATS_IN_CAT_MANIA && j > 0) {
			i = rngRandom(tg->rng, j);
			hatch = hatches + hatchIndices[i];
			critter = newCritter(hatch, CAT);
			addCritter(game, critter);
			free(critter);
			hatchIndices[i] = hatchIndices[--j];
			cats++;
		}
	}
	
	advanceCatsTick(game);
	return 0;
}

//The generator function used for hatches during mouse mania
char mouseManiaGenerator(RNG *rng) {
	static const short SUPER_PROBABILITY_10000 =
		3 * (NORMAL_SUPER_PROBABILITY_10000 / 2);
	static const short SPECIAL_PROBABILITY_10000 =
		NORMAL_SPECIAL_PROBABILITY_10000 / 3;
	
	short r;
	
	if (rngRandom(rng, 10000) <= NORMAL_HATCH_RATE_10000 * 3) {
		r = rngRandom(rng, 10000);
		if (r <= SUPER_PROBABILITY_10000)
			return SUPER_MOUSE;
		else if (r <= SUPER_PROBABILITY_10000 + SPECIAL_PROBABILITY_10000)
			return SPECIAL_MOUSE;
		else
			return MOUSE;
	}
	else
		return NO_CRITTER;
}

char advanceEventMouseMania(TimedGame *tg) {
	Game *game = tg->game;
	hatchTick(game, mouseManiaGenerator);
	advanceCatsTick(game);
	return advanceMiceTick(game);
}

//Increases the speed of the game for the "Speed Up" special event
void speedUp(TimedGame *tg) {
	setSpeed100(tg, 250);
}

//Resets the speed to normal
void resetSpeed(TimedGame *tg) {
	setSpeed100(tg, 100);
}

//Makes the opponent's arrows invisible
void hideOpponentsArrows(TimedGame *tg) {
	tg->drawOpponentsArrows = 0;
}

//Makes the opponent's arrows visible
void showOpponentsArrows(TimedGame *tg) {
	tg->drawOpponentsArrows = 1;
}

//Reverses the positions of the hatches
void switchPlaces(TimedGame *tg) {
	Game *game = tg->game;
	SpaceData *boardData = game->boardData;
	char rocketAPos = game->board->rocketAPos;
	char rocketBPos = game->board->rocketBPos;
	
	if (boardData[rocketAPos].type == SPACE_ROCKET1) {
		boardData[rocketAPos].type = SPACE_ROCKET2;
		boardData[rocketBPos].type = SPACE_ROCKET1;
	}
	else {
		boardData[rocketAPos].type = SPACE_ROCKET1;
		boardData[rocketBPos].type = SPACE_ROCKET2;
	}
}

//Data for the cat attack special event
typedef struct {
	//The cat that is drawn over the appropriate rocket to indicate that that
	//player is losing points.  This cat isn't actually in play and does not
	//appear in the game's critters array; it's just drawn.
	Critter critter;
	//The number of times advanceEventCatAttack was called
	short count;
} CatAttackData;

void startEventCatAttack(TimedGame *tg) {
	Game *game = tg->game;
	char isPlayer1 = (tg->firstEventTick == 1);
	CatAttackData *cad;
	Critter *critter;
	
	tg->event.data = cad = (CatAttackData *)malloc_throw(sizeof(CatAttackData));
	critter = &(cad->critter);
	//The cat exists, but is not actually in play; it is only used for display
	//purposes
	critter->numCritters = 1;
	critter->type = CAT;
	critter->pos =
		(isPlayer1 ^
		 (game->boardData[game->board->rocketAPos].type == SPACE_ROCKET1)) ?
		game->board->rocketAPos : game->board->rocketBPos;
	critter->portion = 0;
	critter->dir = SOUTH;
	cad->count = 0;
	
	if (isPlayer1)
		game->score2 = scoreCat(game->score2);
	else
		game->score1 = scoreCat(game->score1);
}

char advanceEventCatAttack(TimedGame *tg) {
	CatAttackData *cad = (CatAttackData *)tg->event.data;
	(cad->count)++;
	cad->critter.dir = (cad->count / 4) % 2 ? SOUTH : NORTH;
	return 0;
}

void stopEventCatAttack(TimedGame *tg) {
	free(tg->event.data);
}

void drawCatAttack(void *buffer, TimedGame *tg) {
	drawCritter(buffer, &(((CatAttackData *)tg->event.data)->critter), 0);
	drawCrittersPost(buffer);
}

//Removes all arrows
void clearArrows(TimedGame *tg) {
	short i;
	Game *game = tg->game;
	Arrow *arrow;
	Arrow *arrow2;
	
	for(i = 0, arrow = game->p1Arrows, arrow2 = game->p2Arrows; i < MAX_ARROWS; i++, arrow++, arrow2++) {
		if (arrow->exists)
			destroyArrow(arrow, game);
		if (arrow2->exists)
			destroyArrow(arrow2, game);
	}
}

//The generator function used for hatches during cat nap
char noCatsGenerator(RNG *rng) {
	short r;
	
	if (rngRandom(rng, 10000) <= NORMAL_HATCH_RATE_10000) {
		r = rngRandom(rng, 10000);
		if (r <= NORMAL_SUPER_PROBABILITY_10000)
			return SUPER_MOUSE;
		else if (r <= NORMAL_SUPER_PROBABILITY_10000 +
				 NORMAL_SPECIAL_PROBABILITY_10000)
			return SPECIAL_MOUSE;
		else
			return MOUSE;
	}
	else
		return NO_CRITTER;
}

char advanceEventCatNap(TimedGame *tg) {
	Game *game = tg->game;
	hatchTick(game, noCatsGenerator);
	return advanceMiceTick(game);
}

//Reverses the direction of every critter
void aboutFace(TimedGame *tg) {
	short i = 0;
	
	Game *game = tg->game;
	Critter *critter;
	char pos;
	char opp;
	
	for(i = 0, critter = game->critters; i < MAX_CRITTERS; i++, critter++) {
		if (critter->numCritters) {
			pos = critter->pos;
			opp = opposite(critter->dir);
			if (critter->portion) {
				critter->pos = nextPos(critter);
				critter->dir = opp;
				critter->portion = MAX_PORTION - critter->portion;
			}
			else
				critter->dir = newDirArrow(opp,
										   game->boardData[pos].arrowDir,
										   game->board->walls[pos]);
		}
	}
}

//Transfers control of the opponent's rocket to the player for whom a special
//event was just executed
void stealRocket(TimedGame *tg) {
	Board *board = tg->game->board;
	SpaceData *boardData = tg->game->boardData;
	char rocketsInverted = boardData[board->rocketAPos].type == SPACE_ROCKET2;
	
	boardData[board->rocketAPos].type = boardData[board->rocketBPos].type =
		(tg->firstEventTick == 1) ? SPACE_ROCKET1 : SPACE_ROCKET2;
	tg->event.data = malloc_throw(1);
	*((char *)tg->event.data) = rocketsInverted;
}

//Returns control of the opponent's rocket from the player for whom a special
//event was just executed
void returnRocket(TimedGame *tg) {
	Board *board = tg->game->board;
	SpaceData *boardData = tg->game->boardData;
	char rocketsInverted = *((char *)tg->event.data);
	
	boardData[board->rocketAPos].type =
		rocketsInverted ? SPACE_ROCKET2 : SPACE_ROCKET1;
	boardData[board->rocketBPos].type =
		rocketsInverted ? SPACE_ROCKET1 : SPACE_ROCKET2;
	free(tg->event.data);
}

//The number of special events.  Must be at least 2.
#define NUM_EVENTS 10
//The normal duration of special events
#define DEFAULT_EVENT_DURATION 10000l

//Parallel arrays storing information regarding all special events

//The name of the special event.  Each name may be at most 20 characters.
static const char *EVENT_NAMES[NUM_EVENTS] =
	{"Cat Mania",
	 "Mouse Mania",
	 "Speed Up",
	 "Invisible Arrows",
	 "Switch Places",
	 "Cat Attack",
	 "Place Arrows Again",
	 "Cat Nap",
	 "About Face",
	 "Goal Steal"};
//The durations of the special events
static long EVENT_DURATIONS[NUM_EVENTS] =
	{DEFAULT_EVENT_DURATION,
	 DEFAULT_EVENT_DURATION,
	 DEFAULT_EVENT_DURATION,
	 DEFAULT_EVENT_DURATION,
	 0,
	 1000,
	 3000,
	 DEFAULT_EVENT_DURATION * 3 / 4,
	 0,
	 DEFAULT_EVENT_DURATION / 2};
//Functions to start each special event
static void (*START_EVENTS[NUM_EVENTS])(TimedGame *) =
	{destroyCrittersForTimedGame,
	 destroyCrittersForTimedGame,
	 speedUp,
	 hideOpponentsArrows,
	 switchPlaces,
	 startEventCatAttack,
	 clearArrows,
	 doNothingOnTimedGame,
	 aboutFace,
	 stealRocket};
//Functions to advance the game one tick for each special event
static char (*ADVANCE_EVENTS[NUM_EVENTS])(TimedGame *) =
	{advanceEventCatMania,
	 advanceEventMouseMania,
	 advanceTickNormal,
	 advanceTickNormal,
	 advanceTickNormal,
	 advanceEventCatAttack,
	 doNothingAdvanceEvent,
	 advanceEventCatNap,
	 advanceTickNormal,
	 advanceTickNormal};
//Functions to stop each special event
static void (*STOP_EVENTS[NUM_EVENTS])(TimedGame *) =
	{destroyCrittersForTimedGame,
	 doNothingOnTimedGame,
	 resetSpeed,
	 showOpponentsArrows,
	 doNothingOnTimedGame,
	 stopEventCatAttack,
	 doNothingOnTimedGame,
	 doNothingOnTimedGame,
	 doNothingOnTimedGame,
	 returnRocket};
//Functions to draw extra stuff in the background for each special event
static void (*DRAW_EVENTS[NUM_EVENTS])(void *, TimedGame *) =
	{drawNothing,
	 drawNothing,
	 drawNothing,
	 drawNothing,
	 drawNothing,
	 drawCatAttack,
	 drawNothing,
	 drawNothing,
	 drawNothing,
	 drawNothing};

//For what types of critters motion is paused for each special event, one of
//PAUSE_NONE, PAUSE_CATS, or PAUSE_CATS_AND_MICE.  For critters that are paused,
//0 is used as the tenth of a tick that they have advanced to the next space.
static char EVENTS_PAUSE[NUM_EVENTS] =
	{PAUSE_NONE,
	 PAUSE_NONE,
	 PAUSE_NONE,
	 PAUSE_NONE,
	 PAUSE_NONE,
	 PAUSE_CATS_AND_MICE,
	 PAUSE_CATS_AND_MICE,
	 PAUSE_CATS,
	 PAUSE_NONE,
	 PAUSE_NONE};

//Executes a new special event
void event(TimedGame *tg) {
	Event *e = &(tg->event);
	short eventIndex;
	
	//Pick a random event other than the one that was last executed
	while((eventIndex = rngRandom(tg->rng, NUM_EVENTS)) == tg->lastEventIndex);
	
	tg->lastEventIndex = eventIndex;
	tg->advanceTick = ADVANCE_EVENTS[eventIndex];
	e->start = START_EVENTS[eventIndex];
	e->stop = STOP_EVENTS[eventIndex];
	e->draw = DRAW_EVENTS[eventIndex];
	e->pause = EVENTS_PAUSE[eventIndex];
	e->time = EVENT_DURATIONS[eventIndex];
	e->exists = 1;
	postMessage(tg,
				EVENT_NAMES[eventIndex],
				EVENT_MESSAGE_TIME,
				NORMAL_MESSAGE);
}

//Stops the event for the specified TimedGame, if one is running
void stopEvent(TimedGame *tg) {
	Event *e = &(tg->event);
	if (e->exists) {
		e->stop(tg);
		e->exists = 0;
		tg->advanceTick = advanceTickNormal;
	}
}










//Delays the specified timed game by the specified amount of time, so that the
//next time it is advanced, it will not advance by as much time
void delayTimedGame(TimedGame *tg, long t) {
	tg->lastTime += t;
}

//Updates the specified timed game.  Returns the amount of time by which the
//appropriate TimedGame was advanced.  maxTime indicates the maximum amount of
//game time by which this may advance.  Notably, "game time" excludes time when
//this is displaying a message.
long updateTimedGame(TimedGame *tg, long maxTime) {
	long curTime = time();
	long t;
	long timeLeft;
	GameMessage *message = &(tg->message);
	Game *game = tg->game;
	const char *s;
	Event *e = &(tg->event);
	long timeRemainingInGame;
	long initialTimeLeft;
	
	//Safeguard against infinite loops
	if (timerExpired())
		tg->gameOver = 1;
	if (tg->gameOver)
		return 0;
	
	timeLeft = min(min(curTime - tg->lastTime, maxTime), MAX_UPDATE_INTERVAL);
	
	//timeLeft could be negative after a call to delayTimedGame
	if (timeLeft <= 0)
		return 0;
	
	tg->lastTime = curTime;
	
	if (tg->started) {
		initialTimeLeft = timeLeft;
		timeRemainingInGame = tg->timeRemaining;
		
		while(1) {
			//Update the message
			if (message->isShown) {
				t = min(message->time, timeLeft);
				message->time -= t;
				if (message->type == NORMAL_MESSAGE)
					timeRemainingInGame -= t;
				timeLeft -= t;
				
				if (message->time <= 0) {
					//assertion: message->time == 0
					message->isShown = 0;
					if (message->type == END_MESSAGE) {
						tg->gameOver = 1;
						return initialTimeLeft - timeLeft;
					}
				}
			}
			
			if (timeLeft <= 0 || timeRemainingInGame <= 0)
				break;
			
			if (tg->timeUntilNextTick <= timeLeft) {
				//Advance the game by one tick
				decreaseArrowTimes(tg, tg->timeUntilNextTick);
				
				if (e->exists) {
					if (tg->firstEventTick)
						e->start(tg);
					if (e->time <= 0)
						stopEvent(tg);
					e->time -= tg->delay;
				}
				
				if ((tg->firstEventTick = tg->advanceTick(tg))) {
					stopEvent(tg);
					event(tg);
				}
				
				timeRemainingInGame -= tg->timeUntilNextTick;
				timeLeft -= tg->timeUntilNextTick;
				
				//Note that tg->delay may not be the same as delay, as the delay
				//field may have been changed by the special event (e.g. "Speed
				//Up")
				tg->timeUntilNextTick = tg->delay;
			}
			else {
				timeRemainingInGame -= timeLeft;
				tg->timeUntilNextTick -= timeLeft;
				decreaseArrowTimes(tg, timeLeft);
				
				timeLeft = 0;
			}
		}
		
		if (timeRemainingInGame <= 0 &&
				(message->type != END_MESSAGE || !(message->isShown))) {
			if (game->score1 == game->score2)
				s = "Tie";
			else if ((game->score1 > game->score2) == tg->isPlayer1)
				s = "You win!";
			else
				s = "You lose";
			postMessage(tg,
						s,
						END_MESSAGE_TIME + timeRemainingInGame,
						END_MESSAGE);
			timeRemainingInGame = 0;
		}
		
		#if 0
		//Checks for consistency of tg->game->boardData[x].numCats
		short i;
		char cats[54];
		Critter *critter;
		memset(cats, 0, 54);
		for(i = 0; i < 100; i++) {
			critter = tg->game->critters + i;
			if (critter->numCritters && critter->type == CAT) {
				if (critter->portion < 3)
					cats[critter->pos]++;
				else
					cats[nextPos(critter)]++;
			}
		}
		
		for(i = 0; i < 54; i++) {
			if (tg->game->boardData[i].numCats != cats[i])
				exit(0);
		}
		#endif
		
		tg->timeRemaining = timeRemainingInGame;
		return initialTimeLeft;
	}
	else {
		//The first time around, do nothing, so the initial state of the game
		//can be displayed
		tg->timeUntilNextTick = tg->delay;
		tg->started = 1;
		postMessage(tg, "Get ready", START_MESSAGE_TIME, START_MESSAGE);
		return 0;
	}
}











#endif
